from tests_python import debugger_unittest
import sys
import re
import os

CHECK_BASELINE, CHECK_REGULAR, CHECK_CYTHON, CHECK_FRAME_EVAL = "baseline", "regular", "cython", "frame_eval"

pytest_plugins = [
    str("tests_python.debugger_fixtures"),
]

RUNS = 5


class PerformanceWriterThread(debugger_unittest.AbstractWriterThread):
    CHECK = None

    debugger_unittest.AbstractWriterThread.get_environ  # overrides

    def get_environ(self):
        env = os.environ.copy()
        if self.CHECK == CHECK_BASELINE:
            env["PYTHONPATH"] = r"X:\PyDev.Debugger.baseline"

        elif self.CHECK == CHECK_CYTHON:
            env["PYDEVD_USE_CYTHON"] = "YES"
            env["PYDEVD_USE_FRAME_EVAL"] = "NO"

        elif self.CHECK == CHECK_FRAME_EVAL:
            env["PYDEVD_USE_CYTHON"] = "YES"
            env["PYDEVD_USE_FRAME_EVAL"] = "YES"

        elif self.CHECK == CHECK_REGULAR:
            env["PYDEVD_USE_CYTHON"] = "NO"
            env["PYDEVD_USE_FRAME_EVAL"] = "NO"

        else:
            raise AssertionError("Don't know what to check.")
        return env

    debugger_unittest.AbstractWriterThread.get_pydevd_file  # overrides

    def get_pydevd_file(self):
        if self.CHECK == CHECK_BASELINE:
            return os.path.abspath(os.path.join(r"X:\PyDev.Debugger.baseline", "pydevd.py"))
        dirname = os.path.dirname(__file__)
        dirname = os.path.dirname(dirname)
        return os.path.abspath(os.path.join(dirname, "pydevd.py"))


class CheckDebuggerPerformance(debugger_unittest.DebuggerRunner):
    def get_command_line(self):
        return [sys.executable]

    def _get_time_from_result(self, stdout):
        match = re.search(r"TotalTime>>((\d|\.)+)<<", stdout)
        time_taken = match.group(1)
        return float(time_taken)

    def obtain_results(self, benchmark_name, filename):
        class PerformanceCheck(PerformanceWriterThread):
            TEST_FILE = debugger_unittest._get_debugger_test_file(filename)
            BENCHMARK_NAME = benchmark_name

        writer_thread_class = PerformanceCheck

        runs = RUNS
        all_times = []
        for _ in range(runs):
            stdout_ref = []

            def store_stdout(stdout, stderr):
                stdout_ref.append(stdout)

            with self.check_case(writer_thread_class) as writer:
                writer.additional_output_checks = store_stdout
                yield writer

            assert len(stdout_ref) == 1
            all_times.append(self._get_time_from_result(stdout_ref[0]))
            print("partial for: %s: %.3fs" % (writer_thread_class.BENCHMARK_NAME, all_times[-1]))
        if len(all_times) > 3:
            all_times.remove(min(all_times))
            all_times.remove(max(all_times))
        time_when_debugged = sum(all_times) / float(len(all_times))

        args = self.get_command_line()
        args.append(writer_thread_class.TEST_FILE)
        # regular_time = self._get_time_from_result(self.run_process(args, writer_thread=None))
        # simple_trace_time = self._get_time_from_result(self.run_process(args+['--regular-trace'], writer_thread=None))

        if "SPEEDTIN_AUTHORIZATION_KEY" in os.environ:
            SPEEDTIN_AUTHORIZATION_KEY = os.environ["SPEEDTIN_AUTHORIZATION_KEY"]

            # sys.path.append(r'X:\speedtin\pyspeedtin')
            import pyspeedtin  # If the authorization key is there, pyspeedtin must be available
            import pydevd

            pydevd_cython_project_id, pydevd_pure_python_project_id = 6, 7
            if writer_thread_class.CHECK == CHECK_BASELINE:
                project_ids = (pydevd_cython_project_id, pydevd_pure_python_project_id)
            elif writer_thread_class.CHECK == CHECK_REGULAR:
                project_ids = (pydevd_pure_python_project_id,)
            elif writer_thread_class.CHECK == CHECK_CYTHON:
                project_ids = (pydevd_cython_project_id,)
            else:
                raise AssertionError("Wrong check: %s" % (writer_thread_class.CHECK))
            for project_id in project_ids:
                api = pyspeedtin.PySpeedTinApi(authorization_key=SPEEDTIN_AUTHORIZATION_KEY, project_id=project_id)

                benchmark_name = writer_thread_class.BENCHMARK_NAME

                if writer_thread_class.CHECK == CHECK_BASELINE:
                    version = "0.0.1_baseline"
                    return  # No longer commit the baseline (it's immutable right now).
                else:
                    version = (pydevd.__version__,)

                commit_id, branch, commit_date = api.git_commit_id_branch_and_date_from_path(pydevd.__file__)
                api.add_benchmark(benchmark_name)
                api.add_measurement(
                    benchmark_name,
                    value=time_when_debugged,
                    version=version,
                    released=False,
                    branch=branch,
                    commit_id=commit_id,
                    commit_date=commit_date,
                )
                api.commit()

        self.performance_msg = "%s: %.3fs " % (writer_thread_class.BENCHMARK_NAME, time_when_debugged)

    def method_calls_with_breakpoint(self):
        for writer in self.obtain_results("method_calls_with_breakpoint", "_performance_1.py"):
            writer.write_add_breakpoint(17, "method")
            writer.write_make_initial_run()
            writer.finished_ok = True

        return self.performance_msg

    def method_calls_without_breakpoint(self):
        for writer in self.obtain_results("method_calls_without_breakpoint", "_performance_1.py"):
            writer.write_make_initial_run()
            writer.finished_ok = True

        return self.performance_msg

    def method_calls_with_step_over(self):
        for writer in self.obtain_results("method_calls_with_step_over", "_performance_1.py"):
            writer.write_add_breakpoint(26, None)

            writer.write_make_initial_run()
            hit = writer.wait_for_breakpoint_hit("111")

            writer.write_step_over(hit.thread_id)
            hit = writer.wait_for_breakpoint_hit("108")

            writer.write_run_thread(hit.thread_id)
            writer.finished_ok = True

        return self.performance_msg

    def method_calls_with_exception_breakpoint(self):
        for writer in self.obtain_results("method_calls_with_exception_breakpoint", "_performance_1.py"):
            writer.write_add_exception_breakpoint("ValueError")
            writer.write_make_initial_run()
            writer.finished_ok = True

        return self.performance_msg

    def global_scope_1_with_breakpoint(self):
        for writer in self.obtain_results("global_scope_1_with_breakpoint", "_performance_2.py"):
            writer.write_add_breakpoint(writer.get_line_index_with_content("Breakpoint here"), None)
            writer.write_make_initial_run()
            writer.finished_ok = True

        return self.performance_msg

    def global_scope_2_with_breakpoint(self):
        for writer in self.obtain_results("global_scope_2_with_breakpoint", "_performance_3.py"):
            writer.write_add_breakpoint(17, None)
            writer.write_make_initial_run()
            writer.finished_ok = True

        return self.performance_msg


if __name__ == "__main__":
    # Python 3.10
    # Checking: regular
    # method_calls_with_breakpoint: 0.160s
    # method_calls_without_breakpoint: 0.162s
    # method_calls_with_step_over: 0.160s
    # method_calls_with_exception_breakpoint: 0.159s
    # global_scope_1_with_breakpoint: 0.382s
    # global_scope_2_with_breakpoint: 0.141s
    # Partial profile time: 43.90s
    # Checking: cython
    # method_calls_with_breakpoint: 0.094s
    # method_calls_without_breakpoint: 0.094s
    # method_calls_with_step_over: 0.094s
    # method_calls_with_exception_breakpoint: 0.093s
    # global_scope_1_with_breakpoint: 0.252s
    # global_scope_2_with_breakpoint: 0.143s
    # Partial profile time: 41.39s
    # Checking: frame_eval
    # method_calls_with_breakpoint: 0.095s
    # method_calls_without_breakpoint: 0.094s
    # method_calls_with_step_over: 0.096s
    # method_calls_with_exception_breakpoint: 0.096s
    # global_scope_1_with_breakpoint: 0.253s
    # global_scope_2_with_breakpoint: 0.140s
    # Partial profile time: 41.56s

    # Python 3.12 (sys.monitoring)
    # Checking: regular
    # method_calls_with_breakpoint: 0.012s
    # method_calls_without_breakpoint: 0.014s
    # method_calls_with_step_over: 0.014s
    # method_calls_with_exception_breakpoint: 0.013s
    # global_scope_1_with_breakpoint: 0.058s
    # global_scope_2_with_breakpoint: 0.078s
    # Partial profile time: 42.59s
    # Checking: cython
    # method_calls_with_breakpoint: 0.012s
    # method_calls_without_breakpoint: 0.014s
    # method_calls_with_step_over: 0.014s
    # method_calls_with_exception_breakpoint: 0.013s
    # global_scope_1_with_breakpoint: 0.059s
    # global_scope_2_with_breakpoint: 0.078s
    # Partial profile time: 42.37s

    # Python 3.12 (tracing)
    # Checking: regular
    # method_calls_with_breakpoint: 0.205s
    # method_calls_without_breakpoint: 0.201s
    # method_calls_with_step_over: 0.203s
    # method_calls_with_exception_breakpoint: 0.203s
    # global_scope_1_with_breakpoint: 0.499s
    # global_scope_2_with_breakpoint: 0.137s
    # Partial profile time: 44.55s
    # Checking: cython
    # method_calls_with_breakpoint: 0.112s
    # method_calls_without_breakpoint: 0.113s
    # method_calls_with_step_over: 0.113s
    # method_calls_with_exception_breakpoint: 0.115s
    # global_scope_1_with_breakpoint: 0.304s
    # global_scope_2_with_breakpoint: 0.139s
    # Partial profile time: 43.14s

    debugger_unittest.SHOW_WRITES_AND_READS = False
    debugger_unittest.SHOW_OTHER_DEBUG_INFO = False
    debugger_unittest.SHOW_STDOUT = False

    import time

    start_time = time.time()

    tmpdir = None

    msgs = []
    for check in (
        # CHECK_BASELINE, -- Checks against the version checked out at X:\PyDev.Debugger.baseline.
        CHECK_REGULAR,
        CHECK_CYTHON,
        # CHECK_FRAME_EVAL,
    ):
        partial_time = time.time()
        PerformanceWriterThread.CHECK = check
        msgs.append("Checking: %s" % (check,))
        check_debugger_performance = CheckDebuggerPerformance(tmpdir)
        msgs.append(check_debugger_performance.method_calls_with_breakpoint())
        msgs.append(check_debugger_performance.method_calls_without_breakpoint())
        msgs.append(check_debugger_performance.method_calls_with_step_over())
        msgs.append(check_debugger_performance.method_calls_with_exception_breakpoint())
        msgs.append(check_debugger_performance.global_scope_1_with_breakpoint())
        msgs.append(check_debugger_performance.global_scope_2_with_breakpoint())
        msgs.append("Partial profile time: %.2fs" % (time.time() - partial_time,))

    for msg in msgs:
        print(msg)
    print("TotalTime for profile: %.2fs" % (time.time() - start_time,))



import unittest


def setUpModule():
    raise ValueError("This is an INTENTIONAL value error in setUpModule.")


class SetUpModuleTest(unittest.TestCase):
    def setUp(cls):
        pass

    def test_blank(self):
        pass


if __name__ == "__main__":
    unittest.main()



import unittest


class SampleTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_xxxxxx1(self):
        self.fail("Fail test 2")

    def test_xxxxxx2(self):
        pass

    def test_xxxxxx3(self):
        pass

    def test_xxxxxx4(self):
        pass

    def test_non_unique_name(self):
        print("non unique name ran")


class AnotherSampleTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_1(self):
        pass

    def test_2(self):
        """im a doc string"""
        pass

    def todo_not_tested(self):
        """
        Not there by default!
        """


if __name__ == "__main__":
    #    suite = unittest.makeSuite(SampleTest, 'test')
    #    runner = unittest.TextTestRunner( verbosity=3 )
    #    runner.run(suite)
    unittest.main()



"""
@author Fabio Zadrozny
"""
import sys
import unittest
import socket
import urllib
import pytest
import pycompletionserver


IS_JYTHON = sys.platform.find("java") != -1
DEBUG = 0


def dbg(s):
    if DEBUG:
        sys.stdout.write("TEST %s\n" % s)


@pytest.mark.skipif(not IS_JYTHON, reason="Jython related test")
class TestJython(unittest.TestCase):
    def test_it(self):
        dbg("ok")

    def test_message(self):
        t = pycompletionserver.CompletionServer(0)
        t.exit_process_on_kill = False

        l = []
        l.append(("Def", "description", "args"))
        l.append(("Def1", "description1", "args1"))
        l.append(("Def2", "description2", "args2"))

        msg = t.processor.format_completion_message("test_jyserver.py", l)

        self.assertEqual(
            "@@COMPLETIONS(test_jyserver.py,(Def,description,args),(Def1,description1,args1),(Def2,description2,args2))END@@", msg
        )

        l = []
        l.append(("Def", "desc,,r,,i()ption", ""))
        l.append(("Def(1", "descriptio(n1", ""))
        l.append(("De,f)2", "de,s,c,ription2", ""))
        msg = t.processor.format_completion_message(None, l)
        expected = "@@COMPLETIONS(None,(Def,desc%2C%2Cr%2C%2Ci%28%29ption, ),(Def%281,descriptio%28n1, ),(De%2Cf%292,de%2Cs%2Cc%2Cription2, ))END@@"

        self.assertEqual(expected, msg)

    def test_completion_sockets_and_messages(self):
        dbg("test_completion_sockets_and_messages")
        t, socket = self.create_connections()
        self.socket = socket
        dbg("connections created")

        try:
            # now that we have the connections all set up, check the code completion messages.
            msg = urllib.quote_plus("math")

            toWrite = "@@IMPORTS:%sEND@@" % msg
            dbg("writing" + str(toWrite))
            socket.send(toWrite)  # math completions
            completions = self.read_msg()
            dbg(urllib.unquote_plus(completions))

            start = "@@COMPLETIONS("
            self.assertTrue(completions.startswith(start), "%s DOESNT START WITH %s" % (completions, start))
            self.assertTrue(completions.find("@@COMPLETIONS") != -1)
            self.assertTrue(completions.find("END@@") != -1)

            msg = urllib.quote_plus("__builtin__.str")
            toWrite = "@@IMPORTS:%sEND@@" % msg
            dbg("writing" + str(toWrite))
            socket.send(toWrite)  # math completions
            completions = self.read_msg()
            dbg(urllib.unquote_plus(completions))

            start = "@@COMPLETIONS("
            self.assertTrue(completions.startswith(start), "%s DOESNT START WITH %s" % (completions, start))
            self.assertTrue(completions.find("@@COMPLETIONS") != -1)
            self.assertTrue(completions.find("END@@") != -1)

        finally:
            try:
                self.send_kill_msg(socket)

                while not t.ended:
                    pass  # wait until it receives the message and quits.

                socket.close()
            except:
                pass

    def get_free_port(self):
        from _pydev_bundle.pydev_localhost import get_socket_name

        return get_socket_name(close=True)[1]

    def create_connections(self):
        """
        Creates the connections needed for testing.
        """
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind((pycompletionserver.HOST, 0))
        server.listen(1)  # socket to receive messages.

        from thread import start_new_thread

        t = pycompletionserver.CompletionServer(server.getsockname()[1])
        t.exit_process_on_kill = False

        start_new_thread(t.run, ())

        sock, _addr = server.accept()

        return t, sock

    def read_msg(self):
        msg = "@@PROCESSING_END@@"
        while msg.startswith("@@PROCESSING"):
            msg = self.socket.recv(1024)
            if msg.startswith("@@PROCESSING:"):
                dbg("Status msg:" + str(msg))

        while msg.find("END@@") == -1:
            msg += self.socket.recv(1024)

        return msg

    def send_kill_msg(self, socket):
        socket.send(pycompletionserver.MSG_KILL_SERVER)


# Run for jython in command line:
# c:\bin\jython2.7.0\bin\jython.exe -m py.test tests\test_jyserver.py



import threading
import unittest
import os
import pytest
import pydevconsole

from _pydev_bundle.pydev_imports import xmlrpclib, SimpleXMLRPCServer
from _pydev_bundle.pydev_localhost import get_localhost

try:
    raw_input
    raw_input_name = "raw_input"
except NameError:
    raw_input_name = "input"

try:
    from IPython import core  # @UnusedImport

    has_ipython = True
except:
    has_ipython = False


# =======================================================================================================================
# Test
# =======================================================================================================================
@pytest.mark.skipif(os.environ.get("TRAVIS") == "true" or not has_ipython, reason="Too flaky on Travis (and requires IPython).")
class Test(unittest.TestCase):
    def start_client_thread(self, client_port):
        class ClientThread(threading.Thread):
            def __init__(self, client_port):
                threading.Thread.__init__(self)
                self.client_port = client_port

            def run(self):
                class HandleRequestInput:
                    def RequestInput(self):
                        client_thread.requested_input = True
                        return "RequestInput: OK"

                    def NotifyFinished(self, *args, **kwargs):
                        client_thread.notified_finished += 1
                        return 1

                handle_request_input = HandleRequestInput()

                from _pydev_bundle import pydev_localhost

                self.client_server = client_server = SimpleXMLRPCServer(
                    (pydev_localhost.get_localhost(), self.client_port), logRequests=False
                )
                client_server.register_function(handle_request_input.RequestInput)
                client_server.register_function(handle_request_input.NotifyFinished)
                client_server.serve_forever()

            def shutdown(self):
                return
                self.client_server.shutdown()

        client_thread = ClientThread(client_port)
        client_thread.requested_input = False
        client_thread.notified_finished = 0
        client_thread.daemon = True
        client_thread.start()
        return client_thread

    def get_free_addresses(self):
        from _pydev_bundle.pydev_localhost import get_socket_names

        socket_names = get_socket_names(2, close=True)
        return [socket_name[1] for socket_name in socket_names]

    def test_server(self):
        # Just making sure that the singleton is created in this thread.
        from _pydev_bundle.pydev_ipython_console_011 import get_pydev_frontend

        get_pydev_frontend(get_localhost(), 0)

        client_port, server_port = self.get_free_addresses()

        class ServerThread(threading.Thread):
            def __init__(self, client_port, server_port):
                threading.Thread.__init__(self)
                self.client_port = client_port
                self.server_port = server_port

            def run(self):
                from _pydev_bundle import pydev_localhost

                print("Starting server with:", pydev_localhost.get_localhost(), self.server_port, self.client_port)
                pydevconsole.start_server(pydev_localhost.get_localhost(), self.server_port, self.client_port)

        server_thread = ServerThread(client_port, server_port)
        server_thread.daemon = True
        server_thread.start()

        client_thread = self.start_client_thread(client_port)  # @UnusedVariable

        try:
            import time

            time.sleep(0.3)  # let's give it some time to start the threads

            from _pydev_bundle import pydev_localhost

            server = xmlrpclib.Server("http://%s:%s" % (pydev_localhost.get_localhost(), server_port))
            server.execLine("import sys; print('Running with: %s %s' % (sys.executable or sys.platform, sys.version))")
            server.execLine("class Foo:")
            server.execLine("    pass")
            server.execLine("")
            server.execLine("foo = Foo()")
            server.execLine("a = %s()" % raw_input_name)
            initial = time.time()
            while not client_thread.requested_input:
                if time.time() - initial > 2:
                    raise AssertionError("Did not get the return asked before the timeout.")
                time.sleep(0.1)
            frame_xml = server.getFrame()
            self.assertTrue("RequestInput" in frame_xml, "Did not fid RequestInput in:\n%s" % (frame_xml,))
        finally:
            client_thread.shutdown()



# coding: utf-8
from contextlib import contextmanager
import os

import pytest

from tests_python import debugger_unittest
from tests_python.debugger_unittest import (
    get_free_port,
    overrides,
    IS_CPYTHON,
    IS_JYTHON,
    IS_IRONPYTHON,
    CMD_ADD_DJANGO_EXCEPTION_BREAK,
    CMD_REMOVE_DJANGO_EXCEPTION_BREAK,
    CMD_ADD_EXCEPTION_BREAK,
    wait_for_condition,
    IS_PYPY,
)
from _pydevd_bundle.pydevd_comm_constants import file_system_encoding

import sys
from _pydevd_bundle.pydevd_constants import IS_WINDOWS


def get_java_location():
    from java.lang import System  # @UnresolvedImport

    jre_dir = System.getProperty("java.home")
    for f in [os.path.join(jre_dir, "bin", "java.exe"), os.path.join(jre_dir, "bin", "java")]:
        if os.path.exists(f):
            return f
    raise RuntimeError("Unable to find java executable")


def get_jython_jar():
    from java.lang import ClassLoader  # @UnresolvedImport

    cl = ClassLoader.getSystemClassLoader()
    paths = map(lambda url: url.getFile(), cl.getURLs())
    for p in paths:
        if "jython.jar" in p:
            return p
    raise RuntimeError("Unable to find jython.jar")


class _WriterThreadCaseMSwitch(debugger_unittest.AbstractWriterThread):
    TEST_FILE = "tests_python.resources._debugger_case_m_switch"
    IS_MODULE = True

    @overrides(debugger_unittest.AbstractWriterThread.get_environ)
    def get_environ(self):
        env = os.environ.copy()
        curr_pythonpath = env.get("PYTHONPATH", "")

        root_dirname = os.path.dirname(os.path.dirname(__file__))

        curr_pythonpath += root_dirname + os.pathsep
        env["PYTHONPATH"] = curr_pythonpath
        return env

    @overrides(debugger_unittest.AbstractWriterThread.get_main_filename)
    def get_main_filename(self):
        return debugger_unittest._get_debugger_test_file("_debugger_case_m_switch.py")


class _WriterThreadCaseModuleWithEntryPoint(_WriterThreadCaseMSwitch):
    TEST_FILE = "tests_python.resources._debugger_case_module_entry_point:main"
    IS_MODULE = True

    @overrides(_WriterThreadCaseMSwitch.get_main_filename)
    def get_main_filename(self):
        return debugger_unittest._get_debugger_test_file("_debugger_case_module_entry_point.py")


class AbstractWriterThreadCaseFlask(debugger_unittest.AbstractWriterThread):
    FORCE_KILL_PROCESS_WHEN_FINISHED_OK = True
    FLASK_FOLDER = None

    TEST_FILE = "flask"
    IS_MODULE = True

    def write_add_breakpoint_jinja2(self, line, func, template):
        """
        @param line: starts at 1
        """
        assert self.FLASK_FOLDER is not None
        breakpoint_id = self.next_breakpoint_id()
        template_file = debugger_unittest._get_debugger_test_file(os.path.join(self.FLASK_FOLDER, "templates", template))
        self.write("111\t%s\t%s\t%s\t%s\t%s\t%s\tNone\tNone" % (self.next_seq(), breakpoint_id, "jinja2-line", template_file, line, func))
        self.log.append("write_add_breakpoint_jinja: %s line: %s func: %s" % (breakpoint_id, line, func))
        return breakpoint_id

    def write_add_exception_breakpoint_jinja2(self, exception="jinja2-Exception"):
        self.write("%s\t%s\t%s\t%s\t%s\t%s" % (CMD_ADD_EXCEPTION_BREAK, self.next_seq(), exception, 2, 0, 0))

    @overrides(debugger_unittest.AbstractWriterThread.get_environ)
    def get_environ(self):
        import platform

        env = os.environ.copy()
        env["FLASK_APP"] = "app.py"
        env["FLASK_ENV"] = "development"
        env["FLASK_DEBUG"] = "0"
        if platform.system() != "Windows":
            locale = "en_US.utf8" if platform.system() == "Linux" else "en_US.UTF-8"
            env.update(
                {
                    "LC_ALL": locale,
                    "LANG": locale,
                }
            )
        return env

    def get_cwd(self):
        return debugger_unittest._get_debugger_test_file(self.FLASK_FOLDER)

    def get_command_line_args(self):
        assert self.FLASK_FOLDER is not None
        free_port = get_free_port()
        self.flask_port = free_port
        return [
            "flask",
            "run",
            "--no-debugger",
            "--no-reload",
            "--with-threads",
            "--port",
            str(free_port),
        ]

    def _ignore_stderr_line(self, line):
        if debugger_unittest.AbstractWriterThread._ignore_stderr_line(self, line):
            return True

        if "Running on http:" in line:
            return True

        if "GET / HTTP/" in line:
            return True

        return False

    def create_request_thread(self, url=""):
        return debugger_unittest.AbstractWriterThread.create_request_thread(self, "http://127.0.0.1:%s%s" % (self.flask_port, url))


class AbstractWriterThreadCaseDjango(debugger_unittest.AbstractWriterThread):
    FORCE_KILL_PROCESS_WHEN_FINISHED_OK = True
    DJANGO_FOLDER = None

    def _ignore_stderr_line(self, line):
        if debugger_unittest.AbstractWriterThread._ignore_stderr_line(self, line):
            return True

        if "GET /my_app" in line:
            return True

        return False

    def get_command_line_args(self):
        assert self.DJANGO_FOLDER is not None
        free_port = get_free_port()
        self.django_port = free_port
        return [
            debugger_unittest._get_debugger_test_file(os.path.join(self.DJANGO_FOLDER, "manage.py")),
            "runserver",
            "--noreload",
            "--nothreading",
            str(free_port),
        ]

    def write_add_breakpoint_django(self, line, func, template):
        """
        @param line: starts at 1
        """
        assert self.DJANGO_FOLDER is not None
        breakpoint_id = self.next_breakpoint_id()
        template_file = debugger_unittest._get_debugger_test_file(
            os.path.join(self.DJANGO_FOLDER, "my_app", "templates", "my_app", template)
        )
        self.write("111\t%s\t%s\t%s\t%s\t%s\t%s\tNone\tNone" % (self.next_seq(), breakpoint_id, "django-line", template_file, line, func))
        self.log.append("write_add_django_breakpoint: %s line: %s func: %s" % (breakpoint_id, line, func))
        return breakpoint_id

    def write_add_exception_breakpoint_django(self, exception="Exception"):
        self.write("%s\t%s\t%s" % (CMD_ADD_DJANGO_EXCEPTION_BREAK, self.next_seq(), exception))

    def write_remove_exception_breakpoint_django(self, exception="Exception"):
        self.write("%s\t%s\t%s" % (CMD_REMOVE_DJANGO_EXCEPTION_BREAK, self.next_seq(), exception))

    def create_request_thread(self, url=""):
        return debugger_unittest.AbstractWriterThread.create_request_thread(self, "http://127.0.0.1:%s/%s" % (self.django_port, url))


class DebuggerRunnerSimple(debugger_unittest.DebuggerRunner):
    def get_command_line(self):
        if IS_JYTHON:
            if sys.executable is not None:
                # i.e.: we're running with the provided jython.exe
                return [sys.executable]
            else:
                return [get_java_location(), "-classpath", get_jython_jar(), "org.python.util.jython"]

        if IS_CPYTHON or IS_PYPY:
            return [sys.executable, "-u"]

        if IS_IRONPYTHON:
            return [sys.executable, "-X:Frames"]

        raise RuntimeError("Unable to provide command line")


class DebuggerRunnerRemote(debugger_unittest.DebuggerRunner):
    def get_command_line(self):
        return [sys.executable, "-u"]

    def add_command_line_args(self, args, dap=False):
        writer = self.writer

        ret = args + [self.writer.TEST_FILE]
        ret = writer.update_command_line_args(ret)  # Provide a hook for the writer
        return ret


@pytest.fixture
def debugger_runner_simple(tmpdir):
    return DebuggerRunnerSimple(tmpdir)


@pytest.fixture
def debugger_runner_remote(tmpdir):
    return DebuggerRunnerRemote(tmpdir)


@pytest.fixture
def case_setup(tmpdir, debugger_runner_simple):
    runner = debugger_runner_simple

    class WriterThread(debugger_unittest.AbstractWriterThread):
        pass

    class CaseSetup(object):
        check_non_ascii = False
        NON_ASCII_CHARS = "áéíóú汉字"
        dap = False

        @contextmanager
        def test_file(self, filename, wait_for_port=True, wait_for_initialization=True, **kwargs):
            import shutil

            filename = debugger_unittest._get_debugger_test_file(filename)
            if self.check_non_ascii:
                basedir = str(tmpdir)
                if isinstance(basedir, bytes):
                    basedir = basedir.decode("utf-8")
                if isinstance(filename, bytes):
                    filename = filename.decode("utf-8")

                new_dir = os.path.join(basedir, self.NON_ASCII_CHARS)
                os.makedirs(new_dir)

                new_filename = os.path.join(new_dir, self.NON_ASCII_CHARS + os.path.basename(filename))
                shutil.copyfile(filename, new_filename)
                filename = new_filename

            WriterThread.TEST_FILE = filename
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            with runner.check_case(
                WriterThread, wait_for_port=wait_for_port, wait_for_initialization=wait_for_initialization, dap=self.dap
            ) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_dap(case_setup):
    case_setup.dap = True
    return case_setup


@pytest.fixture
def case_setup_unhandled_exceptions(case_setup):
    original = case_setup.test_file

    def check_test_suceeded_msg(writer, stdout, stderr):
        return "TEST SUCEEDED" in "".join(stderr)

    def additional_output_checks(writer, stdout, stderr):
        # Don't call super as we have an expected exception
        if "ValueError: TEST SUCEEDED" not in stderr:
            raise AssertionError('"ValueError: TEST SUCEEDED" not in stderr.\nstdout:\n%s\n\nstderr:\n%s' % (stdout, stderr))

    def test_file(*args, **kwargs):
        kwargs.setdefault("check_test_suceeded_msg", check_test_suceeded_msg)
        kwargs.setdefault("additional_output_checks", additional_output_checks)
        return original(*args, **kwargs)

    case_setup.test_file = test_file

    return case_setup


@pytest.fixture
def case_setup_remote(debugger_runner_remote):
    class WriterThread(debugger_unittest.AbstractWriterThread):
        pass

    class CaseSetup(object):
        dap = False

        @contextmanager
        def test_file(
            self, filename, wait_for_port=True, access_token=None, client_access_token=None, append_command_line_args=(), **kwargs
        ):
            def update_command_line_args(writer, args):
                ret = debugger_unittest.AbstractWriterThread.update_command_line_args(writer, args)
                wait_for_condition(lambda: hasattr(writer, "port"))
                ret.append(str(writer.port))

                if access_token is not None:
                    ret.append("--access-token")
                    ret.append(access_token)
                if client_access_token is not None:
                    ret.append("--client-access-token")
                    ret.append(client_access_token)

                if self.dap:
                    ret.append("--use-dap-mode")

                ret.extend(append_command_line_args)
                return ret

            WriterThread.TEST_FILE = debugger_unittest._get_debugger_test_file(filename)
            WriterThread.update_command_line_args = update_command_line_args
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            with debugger_runner_remote.check_case(WriterThread, wait_for_port=wait_for_port) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_remote_dap(case_setup_remote):
    case_setup_remote.dap = True
    return case_setup_remote


@pytest.fixture
def case_setup_remote_attach_to_dap(debugger_runner_remote):
    """
    The difference from this to case_setup_remote is that this one will connect to a server
    socket started by the debugger and case_setup_remote will create the server socket and wait
    for a connection from the debugger.
    """

    class WriterThread(debugger_unittest.AbstractWriterThread):
        @overrides(debugger_unittest.AbstractWriterThread.run)
        def run(self):
            # I.e.: don't start socket on start(), rather, the test should call
            # start_socket_client() when needed.
            pass

    class CaseSetup(object):
        dap = True

        @contextmanager
        def test_file(self, filename, port, **kwargs):
            additional_args = kwargs.pop("additional_args", [])

            def update_command_line_args(writer, args):
                ret = debugger_unittest.AbstractWriterThread.update_command_line_args(writer, args)
                ret.append(str(port))
                if self.dap:
                    ret.append("--use-dap-mode")
                ret.extend(additional_args)
                return ret

            WriterThread.TEST_FILE = debugger_unittest._get_debugger_test_file(filename)
            WriterThread.update_command_line_args = update_command_line_args
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            with debugger_runner_remote.check_case(WriterThread, wait_for_port=False) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_multiprocessing(debugger_runner_simple):
    class WriterThread(debugger_unittest.AbstractWriterThread):
        pass

    class CaseSetup(object):
        dap = False

        @contextmanager
        def test_file(self, filename, **kwargs):
            def update_command_line_args(writer, args):
                ret = debugger_unittest.AbstractWriterThread.update_command_line_args(writer, args)
                ret.insert(ret.index("--client"), "--multiprocess")
                if self.dap:
                    ret.insert(ret.index("--client"), "--debug-mode")
                    ret.insert(ret.index("--client"), "debugpy-dap")
                    ret.insert(ret.index("--client"), "--json-dap-http")
                return ret

            WriterThread.update_command_line_args = update_command_line_args
            WriterThread.TEST_FILE = debugger_unittest._get_debugger_test_file(filename)
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            with debugger_runner_simple.check_case(WriterThread) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_multiprocessing_dap(case_setup_multiprocessing):
    case_setup_multiprocessing.dap = True
    return case_setup_multiprocessing


@pytest.fixture
def case_setup_m_switch(debugger_runner_simple):
    class WriterThread(_WriterThreadCaseMSwitch):
        pass

    class CaseSetup(object):
        @contextmanager
        def test_file(self, **kwargs):
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)
            with debugger_runner_simple.check_case(WriterThread) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_m_switch_entry_point(debugger_runner_simple):
    runner = debugger_runner_simple

    class WriterThread(_WriterThreadCaseModuleWithEntryPoint):
        pass

    class CaseSetup(object):
        @contextmanager
        def test_file(self, **kwargs):
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)
            with runner.check_case(WriterThread) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_django(debugger_runner_simple):
    class WriterThread(AbstractWriterThreadCaseDjango):
        pass

    class CaseSetup(object):
        dap = False

        @contextmanager
        def test_file(self, **kwargs):
            import django

            version = [int(x) for x in django.get_version().split(".")][:2]
            if version == [1, 7]:
                django_folder = "my_django_proj_17"
            elif version in ([2, 1], [2, 2], [3, 0], [3, 1], [3, 2], [4, 0], [4, 1], [4, 2]):
                django_folder = "my_django_proj_21"
            else:
                raise AssertionError("Can only check django 1.7 -> 4.2 right now. Found: %s" % (version,))

            WriterThread.DJANGO_FOLDER = django_folder
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            with debugger_runner_simple.check_case(WriterThread, dap=self.dap) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_django_dap(case_setup_django):
    case_setup_django.dap = True
    return case_setup_django


@pytest.fixture
def case_setup_flask(debugger_runner_simple):
    class WriterThread(AbstractWriterThreadCaseFlask):
        pass

    class CaseSetup(object):
        dap = False

        @contextmanager
        def test_file(self, **kwargs):
            WriterThread.FLASK_FOLDER = "flask1"
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            with debugger_runner_simple.check_case(WriterThread, dap=self.dap) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_flask_dap(case_setup_flask):
    case_setup_flask.dap = True
    return case_setup_flask



"""
Django settings for my_django_proj_17 project.

For more information on this file, see
https://docs.djangoproject.com/en/1.7/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.7/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os
BASE_DIR = os.path.dirname(os.path.dirname(__file__))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.7/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'Placeholder_5_sue9bp&j=45#%_hcx3f34k!qnt$mxfd&7zq@7c7t@sn4_l)b'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

TEMPLATE_DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'my_app',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'my_django_proj_17.urls'

WSGI_APPLICATION = 'my_django_proj_17.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.7/ref/settings/#databases

# No database for our test.

# DATABASES = {
#     'default': {
#         'ENGINE': 'django.db.backends.sqlite3',
#         'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
#     }
# }

# Internationalization
# https://docs.djangoproject.com/en/1.7/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.7/howto/static-files/

STATIC_URL = '/static/'



from _pydevd_bundle.pydevd_constants import IS_PY313_OR_GREATER, IS_PY38_OR_GREATER, NULL, IS_PY313_0, IS_PY313_1
from _pydevd_bundle.pydevd_xml import ExceptionOnEvaluate

import sys
import pytest

SOME_LST = ["foo", "bar"]
BAR = "bar"
FOO = "foo"
global_frame = sys._getframe()


def obtain_frame():
    A = 1
    B = 2
    yield sys._getframe()


@pytest.fixture
def disable_critical_log():
    # We want to hide the logging related to _evaluate_with_timeouts not receiving the py_db.
    from _pydev_bundle.pydev_log import log_context
    import io

    stream = io.StringIO()
    with log_context(0, stream):
        yield


@pytest.mark.skipif(IS_PY313_0 or IS_PY313_1, reason="Crashes on Python 3.13.0")
def test_evaluate_expression_basic(disable_critical_log):
    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        evaluate_expression(None, frame, "some_var = 1", is_exec=True)

        assert frame.f_locals["some_var"] == 1

    check(next(iter(obtain_frame())))
    assert "some_var" not in sys._getframe().f_globals

    # as locals == globals, this will also change the current globals
    check(global_frame)
    assert "some_var" in sys._getframe().f_globals
    del sys._getframe().f_globals["some_var"]
    assert "some_var" not in sys._getframe().f_globals


@pytest.mark.skipif(IS_PY313_0 or IS_PY313_1, reason="Crashes on Python 3.13.0")
def test_evaluate_expression_1(disable_critical_log):
    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        eval_txt = """
container = ["abc","efg"]
results = []
for s in container:
    result = [s[i] for i in range(3)]
    results.append(result)
"""
        evaluate_expression(None, frame, eval_txt, is_exec=True)
        assert frame.f_locals["results"] == [["a", "b", "c"], ["e", "f", "g"]]
        assert frame.f_locals["s"] == "efg"

    check(next(iter(obtain_frame())))

    for varname in ["container", "results", "s"]:
        assert varname not in sys._getframe().f_globals

    check(global_frame)
    for varname in ["container", "results", "s"]:
        assert varname in sys._getframe().f_globals

    for varname in ["container", "results", "s"]:
        del sys._getframe().f_globals[varname]


@pytest.mark.skipif(IS_PY313_0 or IS_PY313_1, reason="Crashes on Python 3.13.0")
def test_evaluate_expression_2(disable_critical_log):
    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        eval_txt = "all((x in (BAR, FOO) for x in SOME_LST))"
        assert evaluate_expression(None, frame, eval_txt, is_exec=False)

    check(next(iter(obtain_frame())))
    check(global_frame)


@pytest.mark.skipif(IS_PY313_0 or IS_PY313_1, reason="Crashes on Python 3.13.0")
def test_evaluate_expression_3(disable_critical_log):
    if not IS_PY38_OR_GREATER:
        return

    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        eval_txt = """11 if (some_var := 22) else 33"""
        assert evaluate_expression(None, frame, eval_txt, is_exec=False) == 11

    check(next(iter(obtain_frame())))
    assert "some_var" not in sys._getframe().f_globals

    # as locals == globals, this will also change the current globals
    check(global_frame)
    assert "some_var" in sys._getframe().f_globals
    del sys._getframe().f_globals["some_var"]
    assert "some_var" not in sys._getframe().f_globals


@pytest.mark.skipif(IS_PY313_0 or IS_PY313_1, reason="Crashes on Python 3.13.0")
def test_evaluate_expression_4(disable_critical_log):
    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        eval_txt = """import email;email.foo_value"""
        with pytest.raises(AttributeError):
            evaluate_expression(None, frame, eval_txt, is_exec=True)
        assert "email" in frame.f_locals

    check(next(iter(obtain_frame())))
    assert "email" not in sys._getframe().f_globals

    # as locals == globals, this will also change the current globals
    check(global_frame)
    assert "email" in sys._getframe().f_globals
    del sys._getframe().f_globals["email"]
    assert "email" not in sys._getframe().f_globals


@pytest.mark.skipif(IS_PY313_0 or IS_PY313_1, reason="Crashes on Python 3.13.0")
def test_evaluate_expression_access_globals(disable_critical_log):
    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        eval_txt = """globals()['global_variable'] = 22"""
        evaluate_expression(None, frame, eval_txt, is_exec=True)
        assert "global_variable" not in frame.f_locals
        assert "global_variable" in frame.f_globals

    check(next(iter(obtain_frame())))
    assert "global_variable" in sys._getframe().f_globals
    assert "global_variable" not in sys._getframe().f_locals


@pytest.mark.skipif(IS_PY313_0 or IS_PY313_1, reason="Crashes on Python 3.13.0")
def test_evaluate_expression_create_none(disable_critical_log):
    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        eval_txt = "x = None"
        evaluate_expression(None, frame, eval_txt, is_exec=True)
        assert "x" in frame.f_locals
        assert "x" not in frame.f_globals

    check(next(iter(obtain_frame())))


@pytest.mark.skipif(IS_PY313_0 or IS_PY313_1, reason="Crashes on Python 3.13.0")
def test_evaluate_expression_delete_var(disable_critical_log):
    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        eval_txt = "x = 22"
        evaluate_expression(None, frame, eval_txt, is_exec=True)
        assert "x" in frame.f_locals

        eval_txt = "del x"
        evaluate_expression(None, frame, eval_txt, is_exec=True)
        if IS_PY313_0:
            assert frame.f_locals["x"] == None
        else:
            assert "x" not in frame.f_locals

    check(next(iter(obtain_frame())))


@pytest.mark.skipif(IS_PY313_0 or IS_PY313_1, reason="Crashes on Python 3.13.0")
def test_evaluate_expression_5(disable_critical_log):
    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        eval_txt = "A, B = 5, 6"
        evaluate_expression(None, frame, eval_txt, is_exec=True)
        assert frame.f_locals["A"] == 5
        assert frame.f_locals["B"] == 6

    check(next(iter(obtain_frame())))


class _DummyPyDB(object):
    def __init__(self):
        self.created_pydb_daemon_threads = {}
        self.timeout_tracker = NULL
        self.multi_threads_single_notification = False


try:
    from ast import PyCF_ALLOW_TOP_LEVEL_AWAIT  # @UnusedImport

    CAN_EVALUATE_TOP_LEVEL_ASYNC = True
except:
    CAN_EVALUATE_TOP_LEVEL_ASYNC = False


@pytest.mark.skipif(not CAN_EVALUATE_TOP_LEVEL_ASYNC, reason="Requires top-level async evaluation.")
def test_evaluate_expression_async_exec(disable_critical_log):
    py_db = _DummyPyDB()

    async def async_call(a):
        return a

    async def main():
        from _pydevd_bundle.pydevd_vars import evaluate_expression

        a = 10
        assert async_call is not None  # Make sure it's in the locals.
        frame = sys._getframe()
        eval_txt = "y = await async_call(a)"
        evaluate_expression(py_db, frame, eval_txt, is_exec=True)
        assert frame.f_locals["y"] == a

    import asyncio

    asyncio.run(main())


@pytest.mark.skipif(not CAN_EVALUATE_TOP_LEVEL_ASYNC, reason="Requires top-level async evaluation.")
def test_evaluate_expression_async_exec_as_eval(disable_critical_log):
    py_db = _DummyPyDB()

    async def async_call(a):
        return a

    async def main():
        from _pydevd_bundle.pydevd_vars import evaluate_expression

        assert async_call is not None  # Make sure it's in the locals.
        frame = sys._getframe()
        eval_txt = "await async_call(10)"
        from io import StringIO

        _original_stdout = sys.stdout
        try:
            stringio = sys.stdout = StringIO()
            evaluate_expression(py_db, frame, eval_txt, is_exec=True)
        finally:
            sys.stdout = _original_stdout

        # I.e.: Check that we printed the value obtained in the exec.
        assert "10\n" in stringio.getvalue()

    import asyncio

    asyncio.run(main())


@pytest.mark.skipif(not CAN_EVALUATE_TOP_LEVEL_ASYNC or IS_PY313_0, reason="Requires top-level async evaluation. Crashes on Python 3.13.0")
def test_evaluate_expression_async_exec_error(disable_critical_log):
    py_db = _DummyPyDB()

    async def async_call(a):
        raise RuntimeError("foobar")

    async def main():
        from _pydevd_bundle.pydevd_vars import evaluate_expression

        assert async_call is not None  # Make sure it's in the locals.
        frame = sys._getframe()
        eval_txt = "y = await async_call(10)"
        with pytest.raises(RuntimeError) as e:
            evaluate_expression(py_db, frame, eval_txt, is_exec=True)
            assert "foobar" in str(e)
        assert "y" not in frame.f_locals

    import asyncio

    asyncio.run(main())


@pytest.mark.skipif(not CAN_EVALUATE_TOP_LEVEL_ASYNC or IS_PY313_0, reason="Requires top-level async evaluation. Crashes on Python 3.13.0")
def test_evaluate_expression_async_eval(disable_critical_log):
    py_db = _DummyPyDB()

    async def async_call(a):
        return a

    async def main():
        from _pydevd_bundle.pydevd_vars import evaluate_expression

        a = 10
        assert async_call is not None  # Make sure it's in the locals.
        frame = sys._getframe()
        eval_txt = "await async_call(a)"
        v = evaluate_expression(py_db, frame, eval_txt, is_exec=False)
        if isinstance(v, ExceptionOnEvaluate):
            raise v.result.with_traceback(v.tb)
        assert v == a

    import asyncio

    asyncio.run(main())


@pytest.mark.skipif(not CAN_EVALUATE_TOP_LEVEL_ASYNC or IS_PY313_0, reason="Requires top-level async evaluation. Crashes on Python 3.13.0")
def test_evaluate_expression_async_eval_error(disable_critical_log):
    py_db = _DummyPyDB()

    async def async_call(a):
        raise RuntimeError("foobar")

    async def main():
        from _pydevd_bundle.pydevd_vars import evaluate_expression

        a = 10
        assert async_call is not None  # Make sure it's in the locals.
        frame = sys._getframe()
        eval_txt = "await async_call(a)"
        v = evaluate_expression(py_db, frame, eval_txt, is_exec=False)
        assert isinstance(v, ExceptionOnEvaluate)
        assert "foobar" in str(v.result)

    import asyncio

    asyncio.run(main())


def test_evaluate_expression_name_mangling(disable_critical_log):
    from _pydevd_bundle.pydevd_vars import evaluate_expression

    class SomeObj(object):
        def __init__(self):
            self.__value = 10
            self.frame = sys._getframe()

    obj = SomeObj()
    frame = obj.frame

    eval_txt = """self.__value"""
    v = evaluate_expression(None, frame, eval_txt, is_exec=False)
    if isinstance(v, ExceptionOnEvaluate):
        raise v.result.with_traceback(v.tb)

    assert v == 10



"""
This module should be run to recreate the files that we generate automatically
(i.e.: modules that shouldn't be traced and cython .pyx)
"""

from __future__ import print_function

import os
import struct
import re


def is_python_64bit():
    return struct.calcsize("P") == 8


root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


def get_cython_contents(filename):
    if filename.endswith(".pyc"):
        filename = filename[:-1]

    state = "regular"

    replacements = []

    new_contents = []
    with open(filename, "r") as stream:
        for line in stream:
            strip = line.strip()
            if state == "regular":
                if strip == "# IFDEF CYTHON":
                    state = "cython"

                    new_contents.append(
                        "%s -- DONT EDIT THIS FILE (it is automatically generated)\n" % line.replace("\n", "").replace("\r", "")
                    )
                    continue

                new_contents.append(line)

            elif state == "cython":
                if strip == "# ELSE":
                    state = "nocython"
                    new_contents.append(line)
                    continue

                elif strip == "# ENDIF":
                    state = "regular"
                    new_contents.append(line)
                    continue

                if strip == "#":
                    continue

                assert strip.startswith("# "), 'Line inside # IFDEF CYTHON must start with "# ". Found: %s' % (strip,)
                strip = strip.replace("# ", "", 1).strip()

                if strip.startswith("cython_inline_constant:"):
                    strip = strip.replace("cython_inline_constant:", "")
                    word_to_replace, replacement = strip.split("=")
                    replacements.append((word_to_replace.strip(), replacement.strip()))
                    continue

                line = line.replace("# ", "", 1)
                new_contents.append(line)

            elif state == "nocython":
                if strip == "# ENDIF":
                    state = "regular"
                    new_contents.append(line)
                    continue
                new_contents.append("# %s" % line)

    assert state == "regular", "Error: # IFDEF CYTHON found without # ENDIF"

    ret = "".join(new_contents)

    for word_to_replace, replacement in replacements:
        ret = re.sub(r"\b%s\b" % (word_to_replace,), replacement, ret)

    return ret


def _generate_cython_from_files(target, modules):
    contents = [
        """from __future__ import print_function

# Important: Autogenerated file.

# DO NOT edit manually!
# DO NOT edit manually!
"""
    ]

    found = []
    for mod in modules:
        found.append(mod.__file__)
        contents.append(get_cython_contents(mod.__file__))

    print("Generating cython from: %s" % (found,))

    with open(target, "w") as stream:
        stream.write("".join(contents))


def generate_dont_trace_files():
    template = """# Important: Autogenerated file.

# fmt: off
# DO NOT edit manually!
# DO NOT edit manually!

LIB_FILE = 1
PYDEV_FILE = 2

DONT_TRACE_DIRS = {
%(pydev_dirs)s
}

LIB_FILES_IN_DONT_TRACE_DIRS = {
%(pydev_lib_files_in_dont_trace_dirs)s
}

DONT_TRACE = {
    # commonly used things from the stdlib that we don't want to trace
    'Queue.py':LIB_FILE,
    'queue.py':LIB_FILE,
    'socket.py':LIB_FILE,
    'weakref.py':LIB_FILE,
    '_weakrefset.py':LIB_FILE,
    'linecache.py':LIB_FILE,
    'threading.py':LIB_FILE,
    'dis.py':LIB_FILE,

    # things from pydev that we don't want to trace
%(pydev_files)s
}

# if we try to trace io.py it seems it can get halted (see http://bugs.python.org/issue4716)
DONT_TRACE['io.py'] = LIB_FILE

# Don't trace common encodings too
DONT_TRACE['cp1252.py'] = LIB_FILE
DONT_TRACE['utf_8.py'] = LIB_FILE
DONT_TRACE['codecs.py'] = LIB_FILE

# fmt: on
"""

    pydev_files = []
    pydev_dirs = []
    pydev_lib_files_in_dont_trace_dirs = []

    exclude_dirs = [
        ".git",
        ".settings",
        "build",
        "build_tools",
        "dist",
        "pydevd.egg-info",
        "pydevd_attach_to_process",
        "pydev_sitecustomize",
        "stubs",
        "tests",
        "tests_mainloop",
        "tests_python",
        "tests_runfiles",
        "test_pydevd_reload",
        "third_party",
        "__pycache__",
        "vendored",
        ".mypy_cache",
        "pydevd.egg-info",
    ]

    for root, dirs, files in os.walk(root_dir):
        for d in dirs:
            if d == "pydev_ipython":
                pydev_dirs.append("    '%s': LIB_FILE," % (d,))
                continue

            if "pydev" in d and d != "pydevd.egg-info":
                # print(os.path.join(root, d))
                pydev_dirs.append("    '%s': PYDEV_FILE," % (d,))

        for d in exclude_dirs:
            try:
                dirs.remove(d)
            except:
                pass

        if os.path.basename(root) == "pydev_ipython":
            for f in files:
                if f.endswith(".py"):
                    pydev_lib_files_in_dont_trace_dirs.append("    '%s'," % (f,))
            continue

        for f in files:
            if f.endswith(".py"):
                if f not in (
                    "__init__.py",
                    "runfiles.py",
                    "pydev_coverage.py",
                    "pydev_pysrc.py",
                    "setup.py",
                    "setup_pydevd_cython.py",
                    "interpreterInfo.py",
                    "conftest.py",
                ):
                    pydev_files.append("    '%s': PYDEV_FILE," % (f,))

    contents = template % (
        dict(
            pydev_files="\n".join(sorted(set(pydev_files))),
            pydev_dirs="\n".join(sorted(set(pydev_dirs))),
            pydev_lib_files_in_dont_trace_dirs="\n".join(sorted(set(pydev_lib_files_in_dont_trace_dirs))),
        )
    )
    assert "pydevd.py" in contents
    assert "pydevd_dont_trace.py" in contents
    with open(os.path.join(root_dir, "_pydevd_bundle", "pydevd_dont_trace_files.py"), "w") as stream:
        stream.write(contents)


def remove_if_exists(f):
    try:
        if os.path.exists(f):
            os.remove(f)
    except:
        import traceback

        traceback.print_exc()


def generate_cython_module():
    _generate_cython_module()
    _generate_sys_monitoring_cython_module()


def _generate_cython_module():
    print("Removing pydevd_cython.pyx")
    remove_if_exists(os.path.join(root_dir, "_pydevd_bundle", "pydevd_cython.pyx"))

    target = os.path.join(root_dir, "_pydevd_bundle", "pydevd_cython.pyx")
    curr = os.environ.get("PYDEVD_USE_CYTHON")
    try:
        os.environ["PYDEVD_USE_CYTHON"] = "NO"

        from _pydevd_bundle import pydevd_additional_thread_info_regular
        from _pydevd_bundle import pydevd_frame, pydevd_trace_dispatch_regular

        _generate_cython_from_files(target, [pydevd_additional_thread_info_regular, pydevd_frame, pydevd_trace_dispatch_regular])
    finally:
        if curr is None:
            del os.environ["PYDEVD_USE_CYTHON"]
        else:
            os.environ["PYDEVD_USE_CYTHON"] = curr


def _generate_sys_monitoring_cython_module():
    print("Removing _pydevd_sys_monitoring_cython.pyx")
    remove_if_exists(os.path.join(root_dir, "_pydevd_sys_monitoring", "_pydevd_sys_monitoring_cython.pyx"))

    target = os.path.join(root_dir, "_pydevd_sys_monitoring", "_pydevd_sys_monitoring_cython.pyx")
    curr = os.environ.get("PYDEVD_USE_CYTHON")
    try:
        os.environ["PYDEVD_USE_CYTHON"] = "NO"

        from _pydevd_sys_monitoring import _pydevd_sys_monitoring

        _generate_cython_from_files(target, [_pydevd_sys_monitoring])
    finally:
        if curr is None:
            del os.environ["PYDEVD_USE_CYTHON"]
        else:
            os.environ["PYDEVD_USE_CYTHON"] = curr


if __name__ == "__main__":
    generate_dont_trace_files()
    generate_cython_module()



import sys
import os


def find_in_pythonpath(module_name):
    # Check all the occurrences where we could match the given module/package in the PYTHONPATH.
    #
    # This is a simplistic approach, but probably covers most of the cases we're interested in
    # (i.e.: this may fail in more elaborate cases of import customization or .zip imports, but
    # this should be rare in general).
    found_at = []

    parts = module_name.split(".")  # split because we need to convert mod.name to mod/name
    for path in sys.path:
        target = os.path.join(path, *parts)
        target_py = target + ".py"
        if os.path.isdir(target):
            found_at.append(target)
        if os.path.exists(target_py):
            found_at.append(target_py)
    return found_at


class DebuggerInitializationError(Exception):
    pass


class VerifyShadowedImport(object):
    def __init__(self, import_name):
        self.import_name = import_name

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            if exc_type == DebuggerInitializationError:
                return False  # It's already an error we generated.

            # We couldn't even import it...
            found_at = find_in_pythonpath(self.import_name)

            if len(found_at) <= 1:
                # It wasn't found anywhere or there was just 1 occurrence.
                # Let's just return to show the original error.
                return False

            # We found more than 1 occurrence of the same module in the PYTHONPATH
            # (the user module and the standard library module).
            # Let's notify the user as it seems that the module was shadowed.
            msg = self._generate_shadowed_import_message(found_at)
            raise DebuggerInitializationError(msg)

    def _generate_shadowed_import_message(self, found_at):
        msg = """It was not possible to initialize the debugger due to a module name conflict.

i.e.: the module "%(import_name)s" could not be imported because it is shadowed by:
%(found_at)s
Please rename this file/folder so that the original module from the standard library can be imported.""" % {
            "import_name": self.import_name,
            "found_at": found_at[0],
        }

        return msg

    def check(self, module, expected_attributes):
        msg = ""
        for expected_attribute in expected_attributes:
            try:
                getattr(module, expected_attribute)
            except:
                msg = self._generate_shadowed_import_message([module.__file__])
                break

        if msg:
            raise DebuggerInitializationError(msg)


with VerifyShadowedImport("threading") as verify_shadowed:
    import threading

    verify_shadowed.check(threading, ["Thread", "settrace", "setprofile", "Lock", "RLock", "current_thread"])
    ThreadingEvent = threading.Event
    ThreadingLock = threading.Lock
    threading_current_thread = threading.current_thread

with VerifyShadowedImport("time") as verify_shadowed:
    import time

    verify_shadowed.check(time, ["sleep", "time", "mktime"])

with VerifyShadowedImport("socket") as verify_shadowed:
    import socket

    verify_shadowed.check(socket, ["socket", "gethostname", "getaddrinfo"])

with VerifyShadowedImport("select") as verify_shadowed:
    import select

    verify_shadowed.check(select, ["select"])

with VerifyShadowedImport("code") as verify_shadowed:
    import code as _code

    verify_shadowed.check(_code, ["compile_command", "InteractiveInterpreter"])

with VerifyShadowedImport("_thread") as verify_shadowed:
    import _thread as thread

    verify_shadowed.check(thread, ["start_new_thread", "start_new", "allocate_lock"])

with VerifyShadowedImport("queue") as verify_shadowed:
    import queue as _queue

    verify_shadowed.check(_queue, ["Queue", "LifoQueue", "Empty", "Full", "deque"])

with VerifyShadowedImport("xmlrpclib") as verify_shadowed:
    import xmlrpc.client as xmlrpclib

    verify_shadowed.check(xmlrpclib, ["ServerProxy", "Marshaller", "Server"])

with VerifyShadowedImport("xmlrpc.server") as verify_shadowed:
    import xmlrpc.server as xmlrpcserver

    verify_shadowed.check(xmlrpcserver, ["SimpleXMLRPCServer"])

with VerifyShadowedImport("http.server") as verify_shadowed:
    import http.server as BaseHTTPServer

    verify_shadowed.check(BaseHTTPServer, ["BaseHTTPRequestHandler"])

# If set, this is a version of the threading.enumerate that doesn't have the patching to remove the pydevd threads.
# Note: as it can't be set during execution, don't import the name (import the module and access it through its name).
pydevd_saved_threading_enumerate = None



import subprocess
import sys
import os

target_zip = os.path.join(os.path.dirname(__file__), '_debugger_case_subprocess_target.zip')


def create_zip():
    import zipfile
    z = zipfile.ZipFile(target_zip, 'w', zipfile.ZIP_DEFLATED)
    z.writestr('__main__.py', '''
import _debugger_case_subprocess_zip
_debugger_case_subprocess_zip.call_from_zip()
''')


def call_from_zip():
    print('Called from zip')  # break here


def main():
    env = os.environ.copy()
    env['PYTHONPATH'] = os.path.dirname(__file__)
    result = subprocess.check_output([sys.executable, target_zip], env=env).decode('utf-8')
    assert 'Called from zip' in result
    print('TEST SUCEEDED!')


if __name__ == '__main__':
    if not os.path.exists(target_zip):
        create_zip()
    assert os.path.exists(target_zip)

    main()



import os
import sys
import subprocess


def breaknow():
    print('break here')


if '--fork-in-subprocess' in sys.argv:
    popen = None
    if sys.platform == 'win32':
        popen = subprocess.Popen([sys.executable, __file__, '--forked'])
        pid = popen.pid
    else:
        pid = os.fork()
    try:
        ppid = os.getppid()
    except:
        ppid = '<unknown>'
    print('currently in pid: %s, ppid: %s' % (os.getpid(), ppid))
    print('os.fork returned', pid)
    breaknow()
    # i.e.: wait so that we check for the retcode so that we don't get a traceback such as the one
    # below (as that code in __del__ will only be called if returncode is None).
    # Traceback (most recent call last):
    # File "C:\hostedtoolcache\windows\Python\3.9.13\x64\lib\subprocess.py", line 1055, in __del__
    # self._internal_poll(_deadstate=_maxsize)
    # File "C:\hostedtoolcache\windows\Python\3.9.13\x64\lib\subprocess.py", line 1457, in _internal_poll
    # if _WaitForSingleObject(self._handle, 0) == _WAIT_OBJECT_0:
    # OSError: [WinError 6] The handle is invalid
    if popen is not None:
        popen.wait(20)

elif '--forked' in sys.argv:
    try:
        ppid = os.getppid()
    except:
        ppid = '<unknown>'
    print('currently in pid: %s, ppid: %s' % (os.getpid(), ppid))
    breaknow()

elif '--fork-in-subprocess' not in sys.argv:
    out = subprocess.check_output([sys.executable, __file__, '--fork-in-subprocess'])
    breaknow()
    print('\n\nin pid %s, output from subprocess.run:\n%s' % (os.getpid(), out.decode('utf-8')))
    print('TEST SUCEEDED!')



# coding: utf-8
import os.path
from _pydevd_bundle.pydevd_constants import IS_WINDOWS, IS_MAC
import io
from _pydev_bundle.pydev_log import log_context
import pytest
import sys


@pytest.fixture(autouse=True)
def _reset_ide_os():
    yield
    from pydevd_file_utils import set_ide_os

    set_ide_os("WINDOWS" if sys.platform == "win32" else "UNIX")


@pytest.mark.skipif(sys.platform != "win32", reason="Windows-only test.")
def test_get_path_with_real_case_windows_unc_path(monkeypatch):
    import pydevd_file_utils
    from pydevd_file_utils import get_path_with_real_case

    def temp_listdir(d):
        # When we have a UNC drive in windows the "drive" is something as:
        # \\MACHINE_NAME\MOUNT_POINT\
        if d == "\\\\A\\B\\":
            return ["Cc"]
        raise AssertionError("Unexpected: %s" % (d,))

    monkeypatch.setattr(pydevd_file_utils, "os_path_exists", lambda *args: True)
    monkeypatch.setattr(pydevd_file_utils, "os_listdir", temp_listdir)
    assert get_path_with_real_case(r"\\a\b\cc") == r"\\A\B\Cc"


@pytest.mark.skipif(sys.platform != "win32", reason="Windows-only test.")
def test_get_path_with_real_case_windows_slashes_drive(tmpdir):
    from pydevd_file_utils import get_path_with_real_case

    test_dir = str(tmpdir.mkdir("Test_Convert_Utilities")).lower()
    real_case = get_path_with_real_case(test_dir)
    assert real_case.endswith("Test_Convert_Utilities")

    prefix = "\\\\?\\"
    path = prefix + test_dir
    real_case = get_path_with_real_case(path)
    assert real_case.endswith("Test_Convert_Utilities")
    assert path.startswith(prefix)


@pytest.mark.skipif(not IS_MAC, reason="Mac-only test.")
def test_get_path_with_real_case_mac_os(tmpdir):
    from pydevd_file_utils import get_path_with_real_case

    test_dir = str(tmpdir.mkdir("Test_Convert_Utilities")).lower()
    real_case = get_path_with_real_case(test_dir)
    assert real_case.endswith("Test_Convert_Utilities")


@pytest.mark.skipif(not IS_MAC, reason="Mac-only test.")
def test_double_slash_mac(monkeypatch):
    import pydevd_file_utils
    from pydevd_file_utils import get_path_with_real_case

    def temp_listdir(d):
        if d == "//":
            return ["A"]
        if d == "//A":
            return ["Bb"]
        raise AssertionError("Unexpected: %s" % (d,))

    monkeypatch.setattr(pydevd_file_utils, "os_path_exists", lambda *args: True)
    monkeypatch.setattr(pydevd_file_utils, "os_listdir", temp_listdir)
    assert get_path_with_real_case(r"//a/bb") == r"//A/Bb"


def test_convert_utilities(tmpdir):
    import pydevd_file_utils

    test_dir = str(tmpdir.mkdir("Test_Convert_Utilities"))

    if IS_WINDOWS:
        normalized = pydevd_file_utils.normcase(test_dir)
        assert isinstance(normalized, str)
        assert normalized.lower() == normalized

        upper_version = os.path.join(test_dir, "ÁÉÍÓÚ")
        with open(upper_version, "w") as stream:
            stream.write("test")

        with open(upper_version, "r") as stream:
            assert stream.read() == "test"

        with open(pydevd_file_utils.normcase(upper_version), "r") as stream:
            assert stream.read() == "test"

        assert "~" not in normalized

        for i in range(3):  # Check if cache is ok.
            if i == 2:
                pydevd_file_utils._listdir_cache.clear()

            assert pydevd_file_utils.get_path_with_real_case("<does not EXIST>") == "<does not EXIST>"
            real_case = pydevd_file_utils.get_path_with_real_case(normalized)
            assert isinstance(real_case, str)
            # Note test_dir itself cannot be compared with because pytest may
            # have passed the case normalized.
            assert real_case.endswith("Test_Convert_Utilities")

            if i == 2:
                # Check that we have the expected paths in the cache.
                assert pydevd_file_utils._listdir_cache[os.path.dirname(normalized).lower()] == ["Test_Convert_Utilities"]
                assert (
                    pydevd_file_utils._listdir_cache[(os.path.dirname(normalized).lower(), "Test_Convert_Utilities".lower())] == real_case
                )

        # Check that it works with a shortened path.
        shortened = pydevd_file_utils.convert_to_short_pathname(normalized)
        assert "~" in shortened
        with_real_case = pydevd_file_utils.get_path_with_real_case(shortened)
        assert with_real_case.endswith("Test_Convert_Utilities")
        assert "~" not in with_real_case

    elif IS_MAC:
        assert pydevd_file_utils.normcase(test_dir) == test_dir.lower()
        assert pydevd_file_utils.get_path_with_real_case(test_dir) == test_dir

    else:
        # On Linux, nothing should change
        assert pydevd_file_utils.normcase(test_dir) == test_dir
        assert pydevd_file_utils.get_path_with_real_case(test_dir) == test_dir


def test_source_reference(tmpdir):
    import pydevd_file_utils

    pydevd_file_utils.set_ide_os("WINDOWS")
    if IS_WINDOWS:
        # Client and server are on windows.
        pydevd_file_utils.setup_client_server_paths([("c:\\foo", "c:\\bar")])

        assert pydevd_file_utils.map_file_to_client("c:\\bar\\my") == ("c:\\foo\\my", True)
        assert pydevd_file_utils.get_client_filename_source_reference("c:\\foo\\my") == 0

        assert pydevd_file_utils.map_file_to_client("c:\\another\\my") == ("c:\\another\\my", False)
        source_reference = pydevd_file_utils.get_client_filename_source_reference("c:\\another\\my")
        assert source_reference != 0
        assert pydevd_file_utils.get_server_filename_from_source_reference(source_reference) == "c:\\another\\my"

    else:
        # Client on windows and server on unix
        pydevd_file_utils.set_ide_os("WINDOWS")

        pydevd_file_utils.setup_client_server_paths([("C:\\foo", "/bar")])

        assert pydevd_file_utils.map_file_to_client("/bar/my") == ("C:\\foo\\my", True)
        assert pydevd_file_utils.get_client_filename_source_reference("C:\\foo\\my") == 0

        assert pydevd_file_utils.map_file_to_client("/another/my") == ("\\another\\my", False)
        source_reference = pydevd_file_utils.get_client_filename_source_reference("\\another\\my")
        assert source_reference != 0
        assert pydevd_file_utils.get_server_filename_from_source_reference(source_reference) == "/another/my"

        # Allow relative paths in the server path
        pydevd_file_utils.setup_client_server_paths([("C:\\foo", "./bar")])
        root = os.getcwd()
        server_path = os.path.join(root, "bar/my")

        assert pydevd_file_utils.map_file_to_client(server_path) == ("C:\\foo\\my", True)
        assert pydevd_file_utils.get_client_filename_source_reference("C:\\foo\\my") == 0
        assert pydevd_file_utils.map_file_to_server("c:\\foo\\my") == server_path


@pytest.mark.skipif(sys.platform != "win32", reason="Windows-only test.")
def test_translate_only_drive():
    import pydevd_file_utils

    assert pydevd_file_utils.get_path_with_real_case("c:\\") == "C:\\"


def test_to_server_and_to_client(tmpdir):
    try:

        def check(obtained, expected):
            assert obtained == expected, "%s (%s) != %s (%s)" % (obtained, type(obtained), expected, type(expected))
            if isinstance(obtained, tuple):
                assert isinstance(obtained[0], str)
            else:
                assert isinstance(obtained, str)

            if isinstance(expected, tuple):
                assert isinstance(expected[0], str)
            else:
                assert isinstance(expected, str)

        import pydevd_file_utils

        if IS_WINDOWS:
            # Check with made-up files

            pydevd_file_utils.setup_client_server_paths([("c:\\foo", "c:\\bar"), ("c:\\foo2", "c:\\bar2")])

            stream = io.StringIO()
            with log_context(0, stream=stream):
                pydevd_file_utils.map_file_to_server("y:\\only_exists_in_client_not_in_server")
            assert (
                r'pydev debugger: unable to find translation for: "y:\only_exists_in_client_not_in_server" in ["c:\foo\", "c:\foo2\", "c:\foo", "c:\foo2"] (please revise your path mappings).'
                in stream.getvalue()
            )

            # Client and server are on windows.
            pydevd_file_utils.set_ide_os("WINDOWS")
            for in_eclipse, in_python in [
                ("c:\\foo", "c:\\bar"),
                ("c:/foo", "c:\\bar"),
                ("c:\\foo", "c:/bar"),
                ("c:\\foo", "c:\\bar\\"),
                ("c:/foo", "c:\\bar\\"),
                ("c:\\foo", "c:/bar/"),
                ("c:\\foo\\", "c:\\bar"),
                ("c:/foo/", "c:\\bar"),
                ("c:\\foo\\", "c:/bar"),
            ]:
                PATHS_FROM_ECLIPSE_TO_PYTHON = [(in_eclipse, in_python)]
                pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
                check(pydevd_file_utils.map_file_to_server("c:\\foo\\my"), "c:\\bar\\my")
                check(pydevd_file_utils.map_file_to_server("c:/foo/my"), "c:\\bar\\my")
                check(pydevd_file_utils.map_file_to_server("c:/foo/my/"), "c:\\bar\\my")
                check(pydevd_file_utils.map_file_to_server("c:\\foo\\áéíóú".upper()), "c:\\bar" + "\\áéíóú".upper())
                check(pydevd_file_utils.map_file_to_client("c:\\bar\\my"), ("c:\\foo\\my", True))

            # Client on unix and server on windows
            pydevd_file_utils.set_ide_os("UNIX")
            for in_eclipse, in_python in [
                ("/foo", "c:\\bar"),
                ("/foo", "c:/bar"),
                ("/foo", "c:\\bar\\"),
                ("/foo", "c:/bar/"),
                ("/foo/", "c:\\bar"),
                ("/foo/", "c:\\bar\\"),
            ]:
                PATHS_FROM_ECLIPSE_TO_PYTHON = [(in_eclipse, in_python)]
                pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
                check(pydevd_file_utils.map_file_to_server("/foo/my"), "c:\\bar\\my")
                check(pydevd_file_utils.map_file_to_client("c:\\bar\\my"), ("/foo/my", True))
                check(pydevd_file_utils.map_file_to_client("c:\\bar\\my\\"), ("/foo/my", True))
                check(pydevd_file_utils.map_file_to_client("c:/bar/my"), ("/foo/my", True))
                check(pydevd_file_utils.map_file_to_client("c:/bar/my/"), ("/foo/my", True))

            # Test with 'real' files
            # Client and server are on windows.
            pydevd_file_utils.set_ide_os("WINDOWS")

            test_dir = pydevd_file_utils.get_path_with_real_case(str(tmpdir.mkdir("Foo")))
            os.makedirs(os.path.join(test_dir, "Another"))

            in_eclipse = os.path.join(os.path.dirname(test_dir), "Bar")
            in_python = test_dir
            PATHS_FROM_ECLIPSE_TO_PYTHON = [(in_eclipse, in_python)]
            pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)

            if pydevd_file_utils.map_file_to_server(in_eclipse) != in_python.lower():
                raise AssertionError(
                    "%s != %s\ntmpdir:%s\nin_eclipse: %s\nin_python: %s\ntest_dir: %s"
                    % (pydevd_file_utils.map_file_to_server(in_eclipse), in_python.lower(), tmpdir, in_eclipse, in_python, test_dir)
                )

            found_in_eclipse = pydevd_file_utils.map_file_to_client(in_python)[0]
            assert found_in_eclipse.endswith("Bar")

            assert pydevd_file_utils.map_file_to_server(os.path.join(in_eclipse, "another")) == os.path.join(in_python, "another").lower()
            found_in_eclipse = pydevd_file_utils.map_file_to_client(os.path.join(in_python, "another"))[0]
            assert found_in_eclipse.endswith("Bar\\Another")

            # Client on unix and server on windows
            pydevd_file_utils.set_ide_os("UNIX")
            in_eclipse = "/foo"
            in_python = test_dir
            PATHS_FROM_ECLIPSE_TO_PYTHON = [(in_eclipse, in_python)]
            pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
            assert pydevd_file_utils.map_file_to_server("/foo").lower() == in_python.lower()
            assert pydevd_file_utils.map_file_to_client(in_python) == (in_eclipse, True)

            # Test without translation in place (still needs to fix case and separators)
            pydevd_file_utils.set_ide_os("WINDOWS")
            PATHS_FROM_ECLIPSE_TO_PYTHON = []
            pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
            assert pydevd_file_utils.map_file_to_server(test_dir) == test_dir
            assert pydevd_file_utils.map_file_to_client(test_dir.lower())[0].endswith("\\Foo")
        else:
            # Client on windows and server on unix
            pydevd_file_utils.set_ide_os("WINDOWS")

            PATHS_FROM_ECLIPSE_TO_PYTHON = [("c:\\BAR", "/bar")]

            pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
            assert pydevd_file_utils.map_file_to_server("c:\\bar\\my") == "/bar/my"
            assert pydevd_file_utils.map_file_to_client("/bar/my") == ("c:\\BAR\\my", True)

            for in_eclipse, in_python in [
                ("c:\\foo", "/báéíóúr"),
                ("c:/foo", "/báéíóúr"),
                ("c:/foo/", "/báéíóúr"),
                ("c:/foo/", "/báéíóúr/"),
                ("c:\\foo\\", "/báéíóúr/"),
            ]:
                PATHS_FROM_ECLIPSE_TO_PYTHON = [(in_eclipse, in_python)]

                pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
                assert pydevd_file_utils.map_file_to_server("c:\\foo\\my") == "/báéíóúr/my"
                assert pydevd_file_utils.map_file_to_server("C:\\foo\\my") == "/báéíóúr/my"
                assert pydevd_file_utils.map_file_to_server("C:\\foo\\MY") == "/báéíóúr/MY"
                assert pydevd_file_utils.map_file_to_server("C:\\foo\\MY\\") == "/báéíóúr/MY"
                assert pydevd_file_utils.map_file_to_server("c:\\foo\\my\\file.py") == "/báéíóúr/my/file.py"
                assert pydevd_file_utils.map_file_to_server("c:\\foo\\my\\other\\file.py") == "/báéíóúr/my/other/file.py"
                assert pydevd_file_utils.map_file_to_server("c:/foo/my") == "/báéíóúr/my"
                assert pydevd_file_utils.map_file_to_server("c:\\foo\\my\\") == "/báéíóúr/my"
                assert pydevd_file_utils.map_file_to_server("c:/foo/my/") == "/báéíóúr/my"

                assert pydevd_file_utils.map_file_to_client("/báéíóúr/my") == ("c:\\foo\\my", True)
                assert pydevd_file_utils.map_file_to_client("/báéíóúr/my/") == ("c:\\foo\\my", True)

                # Files for which there's no translation have only their separators updated.
                assert pydevd_file_utils.map_file_to_client("/usr/bin/x.py") == ("\\usr\\bin\\x.py", False)
                assert pydevd_file_utils.map_file_to_client("/usr/bin") == ("\\usr\\bin", False)
                assert pydevd_file_utils.map_file_to_client("/usr/bin/") == ("\\usr\\bin", False)
                assert pydevd_file_utils.map_file_to_server("\\usr\\bin") == "/usr/bin"
                assert pydevd_file_utils.map_file_to_server("\\usr\\bin\\") == "/usr/bin"

                # When we have a client file and there'd be no translation, and making it absolute would
                # do something as '$cwd/$file_received' (i.e.: $cwd/c:/another in the case below),
                # warn the user that it's not correct and the path that should be translated instead
                # and don't make it absolute.
                assert pydevd_file_utils.map_file_to_server("c:\\Another") == "c:/Another"

                assert pydevd_file_utils.map_file_to_server("c:/FoO/my/BAR") == "/báéíóúr/my/BAR"
                assert pydevd_file_utils.map_file_to_client("/báéíóúr/my/BAR") == ("c:\\foo\\my\\BAR", True)

            # Client and server on unix
            pydevd_file_utils.set_ide_os("UNIX")
            in_eclipse = "/foo"
            in_python = "/báéíóúr"
            PATHS_FROM_ECLIPSE_TO_PYTHON = [(in_eclipse, in_python)]
            pydevd_file_utils.setup_client_server_paths(PATHS_FROM_ECLIPSE_TO_PYTHON)
            assert pydevd_file_utils.map_file_to_server("/foo/my") == "/báéíóúr/my"
            assert pydevd_file_utils.map_file_to_client("/báéíóúr/my") == ("/foo/my", True)
    finally:
        pydevd_file_utils.setup_client_server_paths([])


def test_relative_paths(tmpdir):
    """
    We need to check that we can deal with relative paths.

    Use cases:
        - Relative path of file that does not exist:
            Use case is a cython-generated module which is generated from a .pyx which
            is not distributed. In this case we need to resolve the file to a library path file.

        - Relative path of a file that exists but not when resolved from the working directory:
            Use case is a cython-generated module which is generated from a .pyx which is
            distributed. In this case we need to resolve to the real file based on the sys.path
            entries.
    """
    import pydevd_file_utils
    import sys

    sys.path.append(str(tmpdir))
    try:
        pydevd_file_utils.NORM_PATHS_AND_BASE_CONTAINER.clear()
        pydevd_file_utils.NORM_PATHS_CONTAINER.clear()
        abs_path = pydevd_file_utils.get_abs_path_real_path_and_base_from_file("my_dir/my_file.pyx")[0]
        assert "site-packages" in abs_path
        assert os.path.normcase(str(tmpdir)) not in abs_path
        assert not pydevd_file_utils.exists("my_dir/my_file.pyx")

        # If the relative file exists when joined with some entry in the PYTHONPATH we'll consider
        # that the relative path points to that absolute path.
        target_dir = os.path.join(str(tmpdir), "my_dir")
        os.makedirs(target_dir)
        with open(os.path.join(target_dir, "my_file.pyx"), "w") as stream:
            stream.write("empty")

        pydevd_file_utils.NORM_PATHS_AND_BASE_CONTAINER.clear()
        pydevd_file_utils.NORM_PATHS_CONTAINER.clear()
        abs_path = pydevd_file_utils.get_abs_path_real_path_and_base_from_file("my_dir/my_file.pyx")[0]
        assert "site-packages" not in abs_path
        assert str(tmpdir) in abs_path
        assert pydevd_file_utils.exists("my_dir/my_file.pyx")
    finally:
        sys.path.remove(str(tmpdir))


def test_zip_paths(tmpdir):
    import pydevd_file_utils
    import sys
    import zipfile

    for i, zip_basename in enumerate(("MY1.zip", "my2.egg!")):
        zipfile_path = str(tmpdir.join(zip_basename))
        zip_file = zipfile.ZipFile(zipfile_path, "w")
        zip_file.writestr("zipped%s/__init__.py" % (i,), "")
        zip_file.writestr("zipped%s/zipped_contents.py" % (i,), "def call_in_zip():\n    return 1")
        zip_file.close()

        sys.path.append(zipfile_path)
        try:
            import importlib
        except ImportError:
            __import__("zipped%s" % (i,))  # Py2.6 does not have importlib
        else:
            importlib.import_module("zipped%s" % (i,))  # Check that it's importable.

        # Check that we can deal with the zip path.
        assert pydevd_file_utils.exists(zipfile_path)
        abspath, realpath, basename = pydevd_file_utils.get_abs_path_real_path_and_base_from_file(zipfile_path)
        if IS_WINDOWS or IS_MAC:
            assert abspath == zipfile_path
            assert basename == zip_basename.lower()
        else:
            assert abspath == zipfile_path
            assert basename == zip_basename

        # Check that we can deal with zip contents.
        for path in [
            zipfile_path + "/zipped%s/__init__.py" % (i,),
            zipfile_path + "/zipped%s/zipped_contents.py" % (i,),
            zipfile_path + "\\zipped%s\\__init__.py" % (i,),
            zipfile_path + "\\zipped%s\\zipped_contents.py" % (i,),
        ]:
            assert pydevd_file_utils.exists(path), "Expected exists to return True for path:\n%s" % (path,)
            abspath, realpath, basename = pydevd_file_utils.get_abs_path_real_path_and_base_from_file(path)
            assert pydevd_file_utils.exists(abspath), "Expected exists to return True for path:\n%s" % (abspath,)
            assert pydevd_file_utils.exists(realpath), "Expected exists to return True for path:\n%s" % (realpath,)

        assert zipfile_path in pydevd_file_utils._ZIP_SEARCH_CACHE, "%s not in %s" % (
            zipfile_path,
            "\n".join(sorted(pydevd_file_utils._ZIP_SEARCH_CACHE.keys())),
        )


def test_source_mapping():
    from _pydevd_bundle.pydevd_source_mapping import SourceMapping, SourceMappingEntry
    from _pydevd_bundle import pydevd_api

    class _DummyPyDB(object):
        source_mapping = SourceMapping()
        api_received_breakpoints = {}
        file_to_id_to_line_breakpoint = {}
        file_to_id_to_plugin_breakpoint = {}
        breakpoints = {}

    source_mapping = _DummyPyDB.source_mapping

    mapping = [
        SourceMappingEntry(line=3, end_line=6, runtime_line=5, runtime_source="<cell1>"),
        SourceMappingEntry(line=10, end_line=11, runtime_line=1, runtime_source="<cell2>"),
    ]

    api = pydevd_api.PyDevdAPI()
    py_db = _DummyPyDB()
    filename = "c:\\temp\\bar.py" if IS_WINDOWS else "/temp/bar.py"
    api.set_source_mapping(py_db, filename, mapping)

    # Map to server
    assert source_mapping.map_to_server(filename, 1) == (filename, 1, False)
    assert source_mapping.map_to_server(filename, 2) == (filename, 2, False)

    assert source_mapping.map_to_server(filename, 3) == ("<cell1>", 5, True)
    assert source_mapping.map_to_server(filename, 4) == ("<cell1>", 6, True)
    assert source_mapping.map_to_server(filename, 5) == ("<cell1>", 7, True)
    assert source_mapping.map_to_server(filename, 6) == ("<cell1>", 8, True)

    assert source_mapping.map_to_server(filename, 7) == (filename, 7, False)

    assert source_mapping.map_to_server(filename, 10) == ("<cell2>", 1, True)
    assert source_mapping.map_to_server(filename, 11) == ("<cell2>", 2, True)

    assert source_mapping.map_to_server(filename, 12) == (filename, 12, False)

    # Map to client
    assert source_mapping.map_to_client(filename, 1) == (filename, 1, False)
    assert source_mapping.map_to_client(filename, 2) == (filename, 2, False)

    assert source_mapping.map_to_client("<cell1>", 5) == (filename, 3, True)
    assert source_mapping.map_to_client("<cell1>", 6) == (filename, 4, True)
    assert source_mapping.map_to_client("<cell1>", 7) == (filename, 5, True)
    assert source_mapping.map_to_client("<cell1>", 8) == (filename, 6, True)

    assert source_mapping.map_to_client(filename, 7) == (filename, 7, False)

    assert source_mapping.map_to_client("<cell2>", 1) == (filename, 10, True)
    assert source_mapping.map_to_client("<cell2>", 2) == (filename, 11, True)

    assert source_mapping.map_to_client(filename, 12) == (filename, 12, False)


@pytest.mark.skipif(IS_WINDOWS, reason="Linux/Mac-only test")
def test_mapping_conflict_to_client():
    import pydevd_file_utils

    path_mappings = []
    for pathMapping in _MAPPING_CONFLICT:
        localRoot = pathMapping.get("localRoot", "")
        remoteRoot = pathMapping.get("remoteRoot", "")
        if (localRoot != "") and (remoteRoot != ""):
            path_mappings.append((localRoot, remoteRoot))

    pydevd_file_utils.setup_client_server_paths(path_mappings)

    assert pydevd_file_utils.map_file_to_client("/opt/pathsomething/foo.py") == ("/var/home/p2/foo.py", True)

    assert pydevd_file_utils.map_file_to_client("/opt/v2/pathsomething/foo.py") == ("/var/home/p4/foo.py", True)

    # This is an odd case, but the user didn't really put a slash in the end,
    # so, it's possible that this is what the user actually wants.
    assert pydevd_file_utils.map_file_to_client("/opt/v2/path_r1/foo.py") == ("/var/home/p3_r1/foo.py", True)

    # The client said both local and remote end with a slash, so, we can only
    # match it with the slash in the end.
    assert pydevd_file_utils.map_file_to_client("/opt/pathsomething_foo.py") == ("/opt/pathsomething_foo.py", False)


_MAPPING_CONFLICT = [
    {"localRoot": "/var/home/p1/", "remoteRoot": "/opt/path/"},
    {"localRoot": "/var/home/p2/", "remoteRoot": "/opt/pathsomething/"},
    {"localRoot": "/var/home/p3", "remoteRoot": "/opt/v2/path"},
    {"localRoot": "/var/home/p4", "remoteRoot": "/opt/v2/pathsomething"},
]


@pytest.mark.skipif(IS_WINDOWS, reason="Linux/Mac-only test")
def test_mapping_conflict_to_server():
    import pydevd_file_utils

    path_mappings = []
    for pathMapping in _MAPPING_CONFLICT_TO_SERVER:
        localRoot = pathMapping.get("localRoot", "")
        remoteRoot = pathMapping.get("remoteRoot", "")
        if (localRoot != "") and (remoteRoot != ""):
            path_mappings.append((localRoot, remoteRoot))

    pydevd_file_utils.setup_client_server_paths(path_mappings)

    assert pydevd_file_utils.map_file_to_server("/opt/pathsomething/foo.py") == "/var/home/p2/foo.py"

    assert pydevd_file_utils.map_file_to_server("/opt/v2/pathsomething/foo.py") == "/var/home/p4/foo.py"

    # This is an odd case, but the user didn't really put a slash in the end,
    # so, it's possible that this is what the user actually wants.
    assert pydevd_file_utils.map_file_to_server("/opt/v2/path_r1/foo.py") == "/var/home/p3_r1/foo.py"

    # The client said both local and remote end with a slash, so, we can only
    # match it with the slash in the end.
    assert pydevd_file_utils.map_file_to_server("/opt/pathsomething_foo.py") == "/opt/pathsomething_foo.py"


_MAPPING_CONFLICT_TO_SERVER = [
    {"remoteRoot": "/var/home/p1/", "localRoot": "/opt/path/"},
    {"remoteRoot": "/var/home/p2/", "localRoot": "/opt/pathsomething/"},
    {"remoteRoot": "/var/home/p3", "localRoot": "/opt/v2/path"},
    {"remoteRoot": "/var/home/p4", "localRoot": "/opt/v2/pathsomething"},
]


def test_get_fullname(tmp_path):
    """Test that get_fullname correctly resolves module names to file paths.

    This is a regression test for the fix that replaced pkgutil.get_loader
    (removed in Python 3.14) with importlib.util.find_spec.
    """
    from pydevd_file_utils import get_fullname

    # Create a temporary module file
    mod_file = tmp_path / "my_test_mod.py"
    mod_file.write_text("x = 1\n")

    # Add tmp_path to sys.path so the module can be found
    sys.path.insert(0, str(tmp_path))
    try:
        result = get_fullname("my_test_mod")
        assert result is not None
        assert result.endswith("my_test_mod.py")

        # Non-existent module should return None
        result = get_fullname("nonexistent_module_xyz_12345")
        assert result is None

        # A stdlib package with __init__.py should be found
        result = get_fullname("json")
        assert result is not None
        assert result.endswith("__init__.py")
    finally:
        sys.path.remove(str(tmp_path))



from _pydevd_bundle.pydevd_constants import IS_PY36_OR_GREATER, GENERATED_LEN_ATTR_NAME
from _pydevd_bundle import pydevd_constants, pydevd_frame_utils
import pytest
import sys


def check_len_entry(len_entry, first_2_params):
    assert len_entry[:2] == first_2_params
    assert callable(len_entry[2])
    assert len_entry[2]("check") == "len(check)"


def test_dict_resolver():
    from _pydevd_bundle.pydevd_resolver import DictResolver

    dict_resolver = DictResolver()
    dct = {(1, 2): 2, "22": 22}
    contents_debug_adapter_protocol = clear_contents_debug_adapter_protocol(dict_resolver.get_contents_debug_adapter_protocol(dct))
    len_entry = contents_debug_adapter_protocol.pop(-1)
    check_len_entry(len_entry, (GENERATED_LEN_ATTR_NAME, 2))
    if IS_PY36_OR_GREATER:
        assert contents_debug_adapter_protocol == [("(1, 2)", 2, "[(1, 2)]"), ("'22'", 22, "['22']")]

    else:
        assert contents_debug_adapter_protocol == [("'22'", 22, "['22']"), ("(1, 2)", 2, "[(1, 2)]")]


def test_dict_resolver_hex():
    from _pydevd_bundle.pydevd_resolver import DictResolver

    dict_resolver = DictResolver()
    dct = {(1, 10, 100): (10000, 100000, 100000)}
    contents_debug_adapter_protocol = clear_contents_debug_adapter_protocol(
        dict_resolver.get_contents_debug_adapter_protocol(dct, fmt={"hex": True})
    )
    len_entry = contents_debug_adapter_protocol.pop(-1)
    check_len_entry(len_entry, (GENERATED_LEN_ATTR_NAME, 1))
    assert contents_debug_adapter_protocol == [
        ("(0x1, 0xa, 0x64)", (10000, 100000, 100000), "[(1, 10, 100)]"),
    ]


def test_object_resolver_simple():
    from _pydevd_bundle.pydevd_resolver import DefaultResolver

    default_resolver = DefaultResolver()

    class MyObject(object):
        def __init__(self):
            self.a = 10
            self.b = 20

    obj = MyObject()
    dictionary = clear_contents_dictionary(default_resolver.get_dictionary(obj))
    assert dictionary == {"a": 10, "b": 20}

    contents_debug_adapter_protocol = clear_contents_debug_adapter_protocol(default_resolver.get_contents_debug_adapter_protocol(obj))
    assert contents_debug_adapter_protocol == [("a", 10, ".a"), ("b", 20, ".b")]


def test_object_resolver_error():
    from _pydevd_bundle.pydevd_resolver import DefaultResolver

    default_resolver = DefaultResolver()

    class MyObject(object):
        def __init__(self):
            self.a = 10

        def __dir__(self):
            return ["a", "b"]

        def __getattribute__(self, attr_name):
            if attr_name == "b":
                raise RuntimeError("unavailable")
            return object.__getattribute__(self, attr_name)

    obj = MyObject()
    dictionary = default_resolver.get_dictionary(obj)
    b_value = dictionary.pop("b")
    assert dictionary == {"a": 10}
    assert 'raise RuntimeError("unavailable")' in b_value

    contents_debug_adapter_protocol = default_resolver.get_contents_debug_adapter_protocol(obj)
    b_value = contents_debug_adapter_protocol.pop(-1)
    assert contents_debug_adapter_protocol == [("a", 10, ".a")]
    assert b_value[0] == "b"
    assert 'raise RuntimeError("unavailable")' in b_value[1]
    assert b_value[2] == ".b"


def test_object_resolver_hasattr_error():
    from _pydevd_bundle.pydevd_resolver import DefaultResolver
    from _pydevd_bundle.pydevd_xml import get_type

    default_resolver = DefaultResolver()

    class MyObject(object):
        def __getattribute__(self, attr_name):
            raise RuntimeError()

    obj = MyObject()
    dictionary = default_resolver.get_dictionary(obj)
    assert dictionary == {}

    _type_object, type_name, _resolver = get_type(obj)
    assert type_name == "MyObject"


def test_object_resolver__dict__non_strings():
    from _pydevd_bundle.pydevd_resolver import DefaultResolver

    default_resolver = DefaultResolver()

    class MyObject(object):
        def __init__(self):
            self.__dict__[(1, 2)] = (3, 4)

    obj = MyObject()
    dictionary = clear_contents_dictionary(default_resolver.get_dictionary(obj))
    assert dictionary == {"(1, 2)": (3, 4)}

    contents_debug_adapter_protocol = clear_contents_debug_adapter_protocol(default_resolver.get_contents_debug_adapter_protocol(obj))
    assert contents_debug_adapter_protocol == [("(1, 2)", (3, 4), ".__dict__[(1, 2)]")]


def test_django_forms_resolver():
    from _pydevd_bundle.pydevd_resolver import DjangoFormResolver

    django_form_resolver = DjangoFormResolver()

    class MyObject(object):
        def __init__(self):
            self.__dict__[(1, 2)] = (3, 4)
            self.__dict__["errors"] = "foo"

    obj = MyObject()

    dictionary = clear_contents_dictionary(django_form_resolver.get_dictionary(obj))
    assert dictionary == {"(1, 2)": (3, 4), "errors": None}

    obj._errors = "bar"
    dictionary = clear_contents_dictionary(django_form_resolver.get_dictionary(obj))
    assert dictionary == {"(1, 2)": (3, 4), "errors": "bar", "_errors": "bar"}


def clear_contents_debug_adapter_protocol(contents_debug_adapter_protocol):
    lst = []
    for x in contents_debug_adapter_protocol:
        if not x[0].startswith("__"):
            if "<built-in method" in str(x[1]) or "<method-wrapper" in str(x[1]) or "<bound method" in str(x[1]):
                continue

            lst.append(x)

    return lst


def clear_contents_dictionary(dictionary):
    dictionary = dictionary.copy()
    for key in list(dictionary):
        if key.startswith("__") or key in ("count", "index"):
            del dictionary[key]
    return dictionary


def test_tuple_resolver():
    from _pydevd_bundle.pydevd_resolver import TupleResolver

    tuple_resolver = TupleResolver()
    fmt = {"hex": True}
    lst = tuple(range(11))
    contents_debug_adapter_protocol = clear_contents_debug_adapter_protocol(tuple_resolver.get_contents_debug_adapter_protocol(lst))
    len_entry = contents_debug_adapter_protocol.pop(-1)
    assert contents_debug_adapter_protocol == [
        ("00", 0, "[0]"),
        ("01", 1, "[1]"),
        ("02", 2, "[2]"),
        ("03", 3, "[3]"),
        ("04", 4, "[4]"),
        ("05", 5, "[5]"),
        ("06", 6, "[6]"),
        ("07", 7, "[7]"),
        ("08", 8, "[8]"),
        ("09", 9, "[9]"),
        ("10", 10, "[10]"),
    ]
    check_len_entry(len_entry, (GENERATED_LEN_ATTR_NAME, 11))

    assert clear_contents_dictionary(tuple_resolver.get_dictionary(lst)) == {
        "00": 0,
        "01": 1,
        "02": 2,
        "03": 3,
        "04": 4,
        "05": 5,
        "06": 6,
        "07": 7,
        "08": 8,
        "09": 9,
        "10": 10,
        GENERATED_LEN_ATTR_NAME: 11,
    }

    lst = tuple(range(17))
    contents_debug_adapter_protocol = clear_contents_debug_adapter_protocol(
        tuple_resolver.get_contents_debug_adapter_protocol(lst, fmt=fmt)
    )
    len_entry = contents_debug_adapter_protocol.pop(-1)
    assert contents_debug_adapter_protocol == [
        ("0x00", 0, "[0]"),
        ("0x01", 1, "[1]"),
        ("0x02", 2, "[2]"),
        ("0x03", 3, "[3]"),
        ("0x04", 4, "[4]"),
        ("0x05", 5, "[5]"),
        ("0x06", 6, "[6]"),
        ("0x07", 7, "[7]"),
        ("0x08", 8, "[8]"),
        ("0x09", 9, "[9]"),
        ("0x0a", 10, "[10]"),
        ("0x0b", 11, "[11]"),
        ("0x0c", 12, "[12]"),
        ("0x0d", 13, "[13]"),
        ("0x0e", 14, "[14]"),
        ("0x0f", 15, "[15]"),
        ("0x10", 16, "[16]"),
    ]
    check_len_entry(len_entry, (GENERATED_LEN_ATTR_NAME, 17))

    assert clear_contents_dictionary(tuple_resolver.get_dictionary(lst, fmt=fmt)) == {
        "0x00": 0,
        "0x01": 1,
        "0x02": 2,
        "0x03": 3,
        "0x04": 4,
        "0x05": 5,
        "0x06": 6,
        "0x07": 7,
        "0x08": 8,
        "0x09": 9,
        "0x0a": 10,
        "0x0b": 11,
        "0x0c": 12,
        "0x0d": 13,
        "0x0e": 14,
        "0x0f": 15,
        "0x10": 16,
        GENERATED_LEN_ATTR_NAME: 17,
    }

    lst = tuple(range(10))
    contents_debug_adapter_protocol = clear_contents_debug_adapter_protocol(tuple_resolver.get_contents_debug_adapter_protocol(lst))
    len_entry = contents_debug_adapter_protocol.pop(-1)
    assert contents_debug_adapter_protocol == [
        ("0", 0, "[0]"),
        ("1", 1, "[1]"),
        ("2", 2, "[2]"),
        ("3", 3, "[3]"),
        ("4", 4, "[4]"),
        ("5", 5, "[5]"),
        ("6", 6, "[6]"),
        ("7", 7, "[7]"),
        ("8", 8, "[8]"),
        ("9", 9, "[9]"),
    ]
    check_len_entry(len_entry, (GENERATED_LEN_ATTR_NAME, 10))

    assert clear_contents_dictionary(tuple_resolver.get_dictionary(lst)) == {
        "0": 0,
        "1": 1,
        "2": 2,
        "3": 3,
        "4": 4,
        "5": 5,
        "6": 6,
        "7": 7,
        "8": 8,
        "9": 9,
        GENERATED_LEN_ATTR_NAME: 10,
    }

    contents_debug_adapter_protocol = clear_contents_debug_adapter_protocol(
        tuple_resolver.get_contents_debug_adapter_protocol(lst, fmt=fmt)
    )
    len_entry = contents_debug_adapter_protocol.pop(-1)
    assert contents_debug_adapter_protocol == [
        ("0x0", 0, "[0]"),
        ("0x1", 1, "[1]"),
        ("0x2", 2, "[2]"),
        ("0x3", 3, "[3]"),
        ("0x4", 4, "[4]"),
        ("0x5", 5, "[5]"),
        ("0x6", 6, "[6]"),
        ("0x7", 7, "[7]"),
        ("0x8", 8, "[8]"),
        ("0x9", 9, "[9]"),
    ]
    check_len_entry(len_entry, (GENERATED_LEN_ATTR_NAME, 10))

    assert clear_contents_dictionary(tuple_resolver.get_dictionary(lst, fmt=fmt)) == {
        "0x0": 0,
        "0x1": 1,
        "0x2": 2,
        "0x3": 3,
        "0x4": 4,
        "0x5": 5,
        "0x6": 6,
        "0x7": 7,
        "0x8": 8,
        "0x9": 9,
        GENERATED_LEN_ATTR_NAME: 10,
    }


def test_tuple_resolver_mixed():
    from _pydevd_bundle.pydevd_resolver import TupleResolver

    tuple_resolver = TupleResolver()

    class CustomTuple(tuple):
        pass

    my_tuple = CustomTuple([1, 2])
    my_tuple.some_value = 10
    contents_debug_adapter_protocol = clear_contents_debug_adapter_protocol(tuple_resolver.get_contents_debug_adapter_protocol(my_tuple))
    len_entry = contents_debug_adapter_protocol.pop(-1)
    check_len_entry(len_entry, (GENERATED_LEN_ATTR_NAME, 2))
    assert contents_debug_adapter_protocol == [
        ("some_value", 10, ".some_value"),
        ("0", 1, "[0]"),
        ("1", 2, "[1]"),
    ]


def test_tuple_resolver_ctypes():
    import ctypes
    from _pydevd_bundle.pydevd_xml import get_type

    array_type = ctypes.c_int32 * 3
    array_value = array_type(1, 2, 3)
    _, _, resolver = get_type(array_value)

    contents_dap = resolver.get_contents_debug_adapter_protocol(array_value)
    # We only care about the array elements and len() here, so remove all preceding entries.
    while True:
        _, _, path = contents_dap[0]
        if path.startswith("["):
            break
        contents_dap.pop(0)

    len_entry = contents_dap.pop(-1)
    assert contents_dap == [
        ("0", 1, "[0]"),
        ("1", 2, "[1]"),
        ("2", 3, "[2]"),
    ]
    check_len_entry(len_entry, (GENERATED_LEN_ATTR_NAME, 3))

    dict_dap = resolver.get_dictionary(array_value)
    # We only care about the array elements and len() here, so remove all other entries.
    dict_dap = {k: v for k, v in dict_dap.items() if k.isdigit() or k == GENERATED_LEN_ATTR_NAME}
    assert dict_dap == {"0": 1, "1": 2, "2": 3, GENERATED_LEN_ATTR_NAME: 3}


def get_tuple_recursive():
    obj = [
        tuple(range(9)),
        [
            tuple(range(5)),
        ],
    ]

    return sys._getframe()


def get_dict_recursive():
    obj = {1: 1, 2: {3: 3, 4: 4, 5: 5, 6: 6, 7: {8: 8, 9: 9, 10: 10}}}

    return sys._getframe()


class _DummyPyDB(object):
    def __init__(self):
        from _pydevd_bundle.pydevd_api import PyDevdAPI

        self.variable_presentation = PyDevdAPI.VariablePresentation()


class _DAPCheckChildVars:
    def __init__(self, data_regression, monkeypatch):
        self.data_regression = data_regression
        self.monkeypatch = monkeypatch

    def check(self, frame, initial_expanded):
        self.monkeypatch.setattr(pydevd_constants, "PYDEVD_CONTAINER_INITIAL_EXPANDED_ITEMS", initial_expanded + 2)
        self.monkeypatch.setattr(pydevd_constants, "PYDEVD_CONTAINER_BUCKET_SIZE", initial_expanded)

        from _pydevd_bundle.pydevd_suspended_frames import SuspendedFramesManager

        suspended_frames_manager = SuspendedFramesManager()
        py_db = _DummyPyDB()

        # Now, let's enable the list packing with less items.
        with suspended_frames_manager.track_frames(py_db) as tracker:
            # : :type tracker: _FramesTracker
            thread_id = "thread1"
            tracker.track(thread_id, pydevd_frame_utils.create_frames_list_from_frame(frame))

            assert suspended_frames_manager.get_thread_id_for_variable_reference(id(frame)) == thread_id

            found = []
            frame_var = suspended_frames_manager.get_variable(id(frame))
            for level, variable in self.collect_all_dap(frame_var.get_child_variable_named("obj")):
                found.append((("    " * level) + variable.name + ": " + str(variable.value)))

            self.data_regression.check(found)

    def collect_all_dap(self, variable, level=0):
        children_variables = variable.get_children_variables()
        for var in children_variables:
            if var.name in ("special variables", "function variables"):
                continue
            yield level, var
            yield from self.collect_all_dap(var, level + 1)


@pytest.fixture
def dap_check_child_vars(data_regression, monkeypatch):
    yield _DAPCheckChildVars(data_regression, monkeypatch)


@pytest.mark.parametrize("initial_expanded", [300, 2])
def test_get_child_variables_multiple_levels_dap(initial_expanded, dap_check_child_vars):
    frame = get_tuple_recursive()
    dap_check_child_vars.check(frame, initial_expanded)


@pytest.mark.parametrize("initial_expanded", [300, 2])
def test_get_child_variables_dict_multiple_levels_dap(initial_expanded, dap_check_child_vars, monkeypatch):
    monkeypatch.setattr(pydevd_constants, "PYDEVD_CONTAINER_RANDOM_ACCESS_MAX_ITEMS", initial_expanded)
    frame = get_dict_recursive()
    dap_check_child_vars.check(frame, initial_expanded)


@pytest.mark.parametrize("initial_expanded", [300, 2])
def test_get_child_variables_multiple_levels_resolver(data_regression, initial_expanded, monkeypatch):
    monkeypatch.setattr(pydevd_constants, "PYDEVD_CONTAINER_INITIAL_EXPANDED_ITEMS", initial_expanded + 2)
    monkeypatch.setattr(pydevd_constants, "PYDEVD_CONTAINER_BUCKET_SIZE", initial_expanded)

    obj = [
        tuple(range(9)),
        [
            tuple(range(5)),
        ],
    ]
    found = []
    for level, key, val in collect_resolver_dictionary(obj):
        found.append((("    " * level) + key + ": " + str(val)))

    data_regression.check(found)


def _skip_key_in_dict(key):
    try:
        int(key)
    except ValueError:
        if "more" in key or "[" in key:
            return False
        return True
    return False


def collect_resolver_dictionary(obj, level=0):
    from _pydevd_bundle.pydevd_xml import get_type

    resolver = get_type(obj)[-1]
    if resolver is None:
        return

    dct = resolver.get_dictionary(obj)
    for key, val in dct.items():
        if _skip_key_in_dict(key):
            continue
        yield level, key, val
        yield from collect_resolver_dictionary(val, level + 1)



import unittest


class Test(unittest.TestCase):
    def testProcessCommandLine(self):
        from _pydevd_bundle.pydevd_command_line_handling import process_command_line, setup_to_argv

        setup = process_command_line(["pydevd.py", "--port", "1", "--save-threading"])
        assert setup["save-threading"]
        assert setup["port"] == 1
        assert not setup["qt-support"]

        argv = setup_to_argv(setup)
        assert argv[0].endswith("pydevd.py") or argv[0].endswith("pydevd$py.class"), "Expected: %s to end with pydevd.py" % (argv[0],)
        argv = argv[1:]
        assert argv == ["--port", "1", "--save-threading"]

    def testProcessCommandLine2(self):
        from _pydevd_bundle.pydevd_command_line_handling import process_command_line, setup_to_argv

        setup = process_command_line(["pydevd.py", "--port", "1", "--qt-support=auto"])
        assert setup["qt-support"] == "auto"

        setup = process_command_line(["pydevd.py", "--port", "1", "--qt-support"])
        assert setup["qt-support"] == "auto"

        setup = process_command_line(["pydevd.py", "--port", "1", "--qt-support=pyqt4"])
        assert setup["qt-support"] == "pyqt4"

        self.assertRaises(ValueError, process_command_line, ["pydevd.py", "--port", "1", "--qt-support=wrong"])



from gevent import monkey
monkey.patch_all()

import subprocess
import sys
import os

if __name__ == "__main__":
    if '-foo' in sys.argv:
        print('foo called')
    else:
        if os.environ.get('CALL_PYTHON_SUB') == '1':
            assert 'foo called' in subprocess.check_output([sys.executable, __file__, '-foo']).decode('utf-8')
        else:
            subprocess.check_output("tput -T xterm-256color bold".split())
        print('TEST SUCEEDED')



import sys

expected_differences = set(
    line.strip()
    for line in r"""
--- a/_pydevd_bundle/pydevd_cython.c
+++ b/_pydevd_bundle/pydevd_cython.c
-  "_pydevd_bundle\\\\pydevd_cython.pyx",
-  "_pydevd_bundle\\\\pydevd_cython.pxd",
+  "_pydevd_bundle/pydevd_cython.pyx",
+  "_pydevd_bundle/pydevd_cython.pxd",

--- a/_pydevd_frame_eval/pydevd_frame_evaluator.c
+++ b/_pydevd_frame_eval/pydevd_frame_evaluator.c
-  "_pydevd_frame_eval\\\\pydevd_frame_evaluator.pyx",
+  "_pydevd_frame_eval/pydevd_frame_evaluator.pyx",
-  "_pydevd_bundle\\\\pydevd_cython.pxd",
+  "_pydevd_bundle/pydevd_cython.pxd",


-  "_pydevd_bundle/pydevd_cython.pyx",
-  "_pydevd_bundle/pydevd_cython.pxd",
+  "_pydevd_bundle\\\\pydevd_cython.pyx",
+  "_pydevd_bundle\\\\pydevd_cython.pxd",

-  "_pydevd_frame_eval/pydevd_frame_evaluator.pyx",
+  "_pydevd_frame_eval\\\\pydevd_frame_evaluator.pyx",
-  "_pydevd_bundle/pydevd_cython.pxd",
+  "_pydevd_bundle\\\\pydevd_cython.pxd",

--- a/_pydevd_sys_monitoring/_pydevd_sys_monitoring_cython.c
+++ b/_pydevd_sys_monitoring/_pydevd_sys_monitoring_cython.c
-  "_pydevd_sys_monitoring\\\\\\\\_pydevd_sys_monitoring_cython.pyx",
+  "_pydevd_sys_monitoring/_pydevd_sys_monitoring_cython.pyx",
-  ".\\\\\\\\_pydevd_bundle\\\\\\\\pydevd_cython.pxd",
+  "./_pydevd_bundle/pydevd_cython.pxd",
-static const char __pyx_k_pydevd_sys_monitoring__pydevd_s[] = "_pydevd_sys_monitoring\\\\_pydevd_sys_monitoring_cython.pyx";
+static const char __pyx_k_pydevd_sys_monitoring__pydevd_s[] = "_pydevd_sys_monitoring/_pydevd_sys_monitoring_cython.pyx";

-  "_pydevd_sys_monitoring\\\\_pydevd_sys_monitoring_cython.pyx",
-  ".\\\\\\\\_pydevd_bundle\\\\pydevd_cython.pxd",
-static const char __pyx_k_pydevd_sys_monitoring__pydevd_s[] = "_pydevd_sys_monitoring\\_pydevd_sys_monitoring_cython.pyx";
-  ".\\\\\\\\_pydevd_bundle\\\\\\\\pydevd_cython.pxd",
-  ".\\\\_pydevd_bundle\\\\pydevd_cython.pxd",

-  "_pydevd_sys_monitoring/_pydevd_sys_monitoring_cython.pyx",
+  "_pydevd_sys_monitoring\\\\\\\\_pydevd_sys_monitoring_cython.pyx",
-  "./_pydevd_bundle/pydevd_cython.pxd",
+  ".\\\\\\\\_pydevd_bundle\\\\\\\\pydevd_cython.pxd",
-static const char __pyx_k_pydevd_sys_monitoring__pydevd_s[] = "_pydevd_sys_monitoring/_pydevd_sys_monitoring_cython.pyx";
+static const char __pyx_k_pydevd_sys_monitoring__pydevd_s[] = "_pydevd_sys_monitoring\\\\_pydevd_sys_monitoring_cython.pyx";

+  "_pydevd_sys_monitoring\\\\_pydevd_sys_monitoring_cython.pyx",
+  ".\\\\_pydevd_bundle\\\\pydevd_cython.pxd",
+static const char __pyx_k_pydevd_sys_monitoring__pydevd_s[] = "_pydevd_sys_monitoring\\_pydevd_sys_monitoring_cython.pyx";
""".splitlines()
    if line.strip()
)


def main():
    import subprocess

    process = subprocess.Popen("git status --porcelain".split(), stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    output, _ = process.communicate()
    if output:
        if sys.version_info[0] > 2:
            output = output.decode("utf-8")

        files = set()
        for line in output.splitlines():
            filename = line[3:]
            files.add(filename.strip())

        files.discard(".travis_install_python_deps.sh")
        files.discard("miniconda.sh")
        files.discard("build_tools/check_no_git_modifications.py")
        found_unexpected = True
        if files:
            found_unexpected = False
            output = subprocess.check_output("git diff".split())
            for line in output.decode("utf-8").splitlines():
                if line.startswith("+") or line.startswith("-"):
                    if line.strip() not in expected_differences:
                        print("Found unexpected: %r" % (line,))
                        found_unexpected = True

        if files and found_unexpected:
            # If there are modifications, show a diff of the modifications and fail the script.
            # (we're mostly interested in modifications to the .c generated files by cython).
            print("Found modifications in git:\n%s " % (output,))
            print("Files: %s" % (files,))
            print("----------- diff -------------")
            subprocess.call("git diff".split())
            print("----------- end diff -------------")
            sys.exit(1)


if __name__ == "__main__":
    main()



import threading
import time


class MyThread(threading.Thread):
    finished = False

    def __init__(self, stop):
        threading.Thread.__init__(self)
        self.stop = stop

    def func_to_stop(self):
        if self.stop:
            stop_now = True  # break on thread
            time.sleep(.1)

    def run(self):
        i = 0
        while not self.finished:
            i += 1
            if self.stop:
                self.func_to_stop()
            a = 1
            b = 2
            c = 3
            if i % 100 == 0:
                time.sleep(.1)


threads = [MyThread(False), MyThread(True)]

for t in threads:
    t.start()

stop_on_main = True  # break on main

for t in threads:
    t.finished = True

for t in threads:
    t.join()
print('TEST SUCEEDED')



r"""
Creating the needed environments for creating the pre-compiled distribution on Windows:

See:

build_tools\pydevd_release_process.txt

for building binaries/release process.
"""

from __future__ import unicode_literals
import os
import subprocess
import sys

miniconda_envs = os.getenv("MINICONDA_ENVS", r"D:\bin\miniconda\envs")

python_installations = [
    r"%s\py38_64\python.exe" % miniconda_envs,
    r"%s\py39_64\python.exe" % miniconda_envs,
    r"%s\py310_64\python.exe" % miniconda_envs,
    r"%s\py311_64\python.exe" % miniconda_envs,
    r"%s\py312_64\python.exe" % miniconda_envs,
    r"%s\py313_64\python.exe" % miniconda_envs,
    r"%s\py314_64\python.exe" % miniconda_envs,
    # See: build_tools\pydevd_release_process.txt when adding a new one
]

root_dir = os.path.dirname(os.path.dirname(__file__))


def list_binaries():
    for f in os.listdir(os.path.join(root_dir, "_pydevd_bundle")):
        if f.endswith(".pyd"):
            yield f


def extract_version(python_install):
    return python_install.split("\\")[-2][2:]


def main():
    from generate_code import generate_dont_trace_files
    from generate_code import generate_cython_module

    # First, make sure that our code is up to date.
    generate_dont_trace_files()
    generate_cython_module()

    for python_install in python_installations:
        assert os.path.exists(python_install), "%s does not exist." % (python_install,)

    from build import remove_binaries

    remove_binaries([".pyd"])

    for f in list_binaries():
        raise AssertionError("Binary not removed: %s" % (f,))

    for i, python_install in enumerate(python_installations):
        print()
        print("*" * 80)
        print("*" * 80)
        print()
        new_name = "pydevd_cython_%s_%s" % (sys.platform, extract_version(python_install))
        args = [
            python_install,
            os.path.join(root_dir, "build_tools", "build.py"),
            "--no-remove-binaries",
            "--target-pyd-name=%s" % new_name,
            "--force-cython",
        ]
        if i != 0:
            args.append("--no-regenerate-files")
        name_frame_eval = "pydevd_frame_evaluator_%s_%s" % (sys.platform, extract_version(python_install))
        args.append("--target-pyd-frame-eval=%s" % name_frame_eval)
        print("Calling: %s" % (" ".join(args)))

        env = os.environ.copy()
        python_exe_dir = os.path.dirname(python_install)
        env["PATH"] = env["PATH"] + ";" + os.path.join(python_exe_dir, "DLLs") + ";" + os.path.join(python_exe_dir, "Library", "bin")
        subprocess.check_call(args, env=env)


if __name__ == "__main__":
    main()

# To run do:
# See:
#
# build_tools\pydevd_release_process.txt
#
# for building binaries/release process.



import inspect

from _pydev_bundle import pydev_log
from _pydevd_bundle.pydevd_comm import CMD_SET_BREAK, CMD_ADD_EXCEPTION_BREAK
from _pydevd_bundle.pydevd_constants import STATE_SUSPEND, DJANGO_SUSPEND, DebugInfoHolder
from _pydevd_bundle.pydevd_frame_utils import add_exception_to_frame, FCode, just_raised, ignore_exception_trace
from pydevd_file_utils import canonical_normalized_path, absolute_path
from _pydevd_bundle.pydevd_api import PyDevdAPI
from pydevd_plugins.pydevd_line_validation import LineBreakpointWithLazyValidation, ValidationInfo
from _pydev_bundle.pydev_override import overrides

IS_DJANGO18 = False
IS_DJANGO19 = False
IS_DJANGO19_OR_HIGHER = False
try:
    import django

    version = django.VERSION
    IS_DJANGO18 = version[0] == 1 and version[1] == 8
    IS_DJANGO19 = version[0] == 1 and version[1] == 9
    IS_DJANGO19_OR_HIGHER = (version[0] == 1 and version[1] >= 9) or version[0] > 1
except:
    pass


class DjangoLineBreakpoint(LineBreakpointWithLazyValidation):
    def __init__(
        self, canonical_normalized_filename, breakpoint_id, line, condition, func_name, expression, hit_condition=None, is_logpoint=False
    ):
        self.canonical_normalized_filename = canonical_normalized_filename
        LineBreakpointWithLazyValidation.__init__(
            self, breakpoint_id, line, condition, func_name, expression, hit_condition=hit_condition, is_logpoint=is_logpoint
        )

    def __str__(self):
        return "DjangoLineBreakpoint: %s-%d" % (self.canonical_normalized_filename, self.line)


class _DjangoValidationInfo(ValidationInfo):
    @overrides(ValidationInfo._collect_valid_lines_in_template_uncached)
    def _collect_valid_lines_in_template_uncached(self, template):
        lines = set()
        for node in self._iternodes(template.nodelist):
            if node.__class__.__name__ in _IGNORE_RENDER_OF_CLASSES:
                continue
            lineno = self._get_lineno(node)
            if lineno is not None:
                lines.add(lineno)
        return lines

    def _get_lineno(self, node):
        if hasattr(node, "token") and hasattr(node.token, "lineno"):
            return node.token.lineno
        return None

    def _iternodes(self, nodelist):
        for node in nodelist:
            yield node

            try:
                children = node.child_nodelists
            except:
                pass
            else:
                for attr in children:
                    nodelist = getattr(node, attr, None)
                    if nodelist:
                        # i.e.: yield from _iternodes(nodelist)
                        for node in self._iternodes(nodelist):
                            yield node


def add_line_breakpoint(
    pydb,
    type,
    canonical_normalized_filename,
    breakpoint_id,
    line,
    condition,
    expression,
    func_name,
    hit_condition=None,
    is_logpoint=False,
    add_breakpoint_result=None,
    on_changed_breakpoint_state=None,
):
    if type == "django-line":
        django_line_breakpoint = DjangoLineBreakpoint(
            canonical_normalized_filename,
            breakpoint_id,
            line,
            condition,
            func_name,
            expression,
            hit_condition=hit_condition,
            is_logpoint=is_logpoint,
        )
        if not hasattr(pydb, "django_breakpoints"):
            _init_plugin_breaks(pydb)

        if IS_DJANGO19_OR_HIGHER:
            add_breakpoint_result.error_code = PyDevdAPI.ADD_BREAKPOINT_LAZY_VALIDATION
            django_line_breakpoint.add_breakpoint_result = add_breakpoint_result
            django_line_breakpoint.on_changed_breakpoint_state = on_changed_breakpoint_state
        else:
            add_breakpoint_result.error_code = PyDevdAPI.ADD_BREAKPOINT_NO_ERROR

        return django_line_breakpoint, pydb.django_breakpoints
    return None


def after_breakpoints_consolidated(py_db, canonical_normalized_filename, id_to_pybreakpoint, file_to_line_to_breakpoints):
    if IS_DJANGO19_OR_HIGHER:
        django_breakpoints_for_file = file_to_line_to_breakpoints.get(canonical_normalized_filename)
        if not django_breakpoints_for_file:
            return

        if not hasattr(py_db, "django_validation_info"):
            _init_plugin_breaks(py_db)

        # In general we validate the breakpoints only when the template is loaded, but if the template
        # was already loaded, we can validate the breakpoints based on the last loaded value.
        py_db.django_validation_info.verify_breakpoints_from_template_cached_lines(
            py_db, canonical_normalized_filename, django_breakpoints_for_file
        )


def add_exception_breakpoint(pydb, type, exception):
    if type == "django":
        if not hasattr(pydb, "django_exception_break"):
            _init_plugin_breaks(pydb)
        pydb.django_exception_break[exception] = True
        return True
    return False


def _init_plugin_breaks(pydb):
    pydb.django_exception_break = {}
    pydb.django_breakpoints = {}

    pydb.django_validation_info = _DjangoValidationInfo()


def remove_exception_breakpoint(pydb, exception_type, exception):
    if exception_type == "django":
        try:
            del pydb.django_exception_break[exception]
            return True
        except KeyError:
            pass
    return False


def remove_all_exception_breakpoints(pydb):
    if hasattr(pydb, "django_exception_break"):
        pydb.django_exception_break = {}
        return True
    return False


def get_breakpoints(pydb, breakpoint_type):
    if breakpoint_type == "django-line":
        return pydb.django_breakpoints
    return None


def _inherits(cls, *names):
    if cls.__name__ in names:
        return True
    inherits_node = False
    for base in inspect.getmro(cls):
        if base.__name__ in names:
            inherits_node = True
            break
    return inherits_node


_IGNORE_RENDER_OF_CLASSES = ("TextNode", "NodeList")


def _is_django_render_call(frame):
    try:
        name = frame.f_code.co_name
        if name != "render":
            return False

        if "self" not in frame.f_locals:
            return False

        cls = frame.f_locals["self"].__class__

        inherits_node = _inherits(cls, "Node")

        if not inherits_node:
            return False

        clsname = cls.__name__
        if IS_DJANGO19:
            # in Django 1.9 we need to save the flag that there is included template
            if clsname == "IncludeNode":
                if "context" in frame.f_locals:
                    context = frame.f_locals["context"]
                    context._has_included_template = True

        return clsname not in _IGNORE_RENDER_OF_CLASSES
    except:
        pydev_log.exception()
        return False


def _is_django_context_get_call(frame):
    try:
        if "self" not in frame.f_locals:
            return False

        cls = frame.f_locals["self"].__class__

        return _inherits(cls, "BaseContext")
    except:
        pydev_log.exception()
        return False


def _is_django_resolve_call(frame):
    try:
        name = frame.f_code.co_name
        if name != "_resolve_lookup":
            return False

        if "self" not in frame.f_locals:
            return False

        cls = frame.f_locals["self"].__class__

        clsname = cls.__name__
        return clsname == "Variable"
    except:
        pydev_log.exception()
        return False


def _is_django_suspended(thread):
    return thread.additional_info.suspend_type == DJANGO_SUSPEND


def suspend_django(py_db, thread, frame, cmd=CMD_SET_BREAK):
    if frame.f_lineno is None:
        return None

    py_db.set_suspend(thread, cmd)
    thread.additional_info.suspend_type = DJANGO_SUSPEND

    return frame


def _find_django_render_frame(frame):
    while frame is not None and not _is_django_render_call(frame):
        frame = frame.f_back

    return frame


# =======================================================================================================================
# Django Frame
# =======================================================================================================================


def _read_file(filename):
    # type: (str) -> str
    f = open(filename, "r", encoding="utf-8", errors="replace")
    s = f.read()
    f.close()
    return s


def _offset_to_line_number(text, offset):
    curLine = 1
    curOffset = 0
    while curOffset < offset:
        if curOffset == len(text):
            return -1
        c = text[curOffset]
        if c == "\n":
            curLine += 1
        elif c == "\r":
            curLine += 1
            if curOffset < len(text) and text[curOffset + 1] == "\n":
                curOffset += 1

        curOffset += 1

    return curLine


def _get_source_django_18_or_lower(frame):
    # This method is usable only for the Django <= 1.8
    try:
        node = frame.f_locals["self"]
        if hasattr(node, "source"):
            return node.source
        else:
            if IS_DJANGO18:
                # The debug setting was changed since Django 1.8
                pydev_log.error_once(
                    "WARNING: Template path is not available. Set the 'debug' option in the OPTIONS of a DjangoTemplates backend."
                )
            else:
                # The debug setting for Django < 1.8
                pydev_log.error_once(
                    "WARNING: Template path is not available. Please set TEMPLATE_DEBUG=True in your settings.py to make "
                    "django template breakpoints working"
                )
            return None

    except:
        pydev_log.exception()
        return None


def _convert_to_str(s):
    return s


def _get_template_original_file_name_from_frame(frame):
    try:
        if IS_DJANGO19:
            # The Node source was removed since Django 1.9
            if "context" in frame.f_locals:
                context = frame.f_locals["context"]
                if hasattr(context, "_has_included_template"):
                    # if there was included template we need to inspect the previous frames and find its name
                    back = frame.f_back
                    while back is not None and frame.f_code.co_name in ("render", "_render"):
                        locals = back.f_locals
                        if "self" in locals:
                            self = locals["self"]
                            if self.__class__.__name__ == "Template" and hasattr(self, "origin") and hasattr(self.origin, "name"):
                                return _convert_to_str(self.origin.name)
                        back = back.f_back
                else:
                    if hasattr(context, "template") and hasattr(context.template, "origin") and hasattr(context.template.origin, "name"):
                        return _convert_to_str(context.template.origin.name)
            return None
        elif IS_DJANGO19_OR_HIGHER:
            # For Django 1.10 and later there is much simpler way to get template name
            if "self" in frame.f_locals:
                self = frame.f_locals["self"]
                if hasattr(self, "origin") and hasattr(self.origin, "name"):
                    return _convert_to_str(self.origin.name)
            return None

        source = _get_source_django_18_or_lower(frame)
        if source is None:
            pydev_log.debug("Source is None\n")
            return None
        fname = _convert_to_str(source[0].name)

        if fname == "<unknown source>":
            pydev_log.debug("Source name is %s\n" % fname)
            return None
        else:
            return fname
    except:
        if DebugInfoHolder.DEBUG_TRACE_LEVEL >= 2:
            pydev_log.exception("Error getting django template filename.")
        return None


def _get_template_line(frame):
    if IS_DJANGO19_OR_HIGHER:
        node = frame.f_locals["self"]
        if hasattr(node, "token") and hasattr(node.token, "lineno"):
            return node.token.lineno
        else:
            return None

    source = _get_source_django_18_or_lower(frame)
    original_filename = _get_template_original_file_name_from_frame(frame)
    if original_filename is not None:
        try:
            absolute_filename = absolute_path(original_filename)
            return _offset_to_line_number(_read_file(absolute_filename), source[1][0])
        except:
            return None
    return None


class DjangoTemplateFrame(object):
    IS_PLUGIN_FRAME = True

    def __init__(self, frame):
        original_filename = _get_template_original_file_name_from_frame(frame)
        self._back_context = frame.f_locals["context"]
        self.f_code = FCode("Django Template", original_filename)
        self.f_lineno = _get_template_line(frame)
        self.f_back = frame
        self.f_globals = {}
        self.f_locals = self._collect_context(self._back_context)
        self.f_trace = None

    def _collect_context(self, context):
        res = {}
        try:
            for d in context.dicts:
                for k, v in d.items():
                    res[k] = v
        except AttributeError:
            pass
        return res

    def _change_variable(self, name, value):
        for d in self._back_context.dicts:
            for k, v in d.items():
                if k == name:
                    d[k] = value


class DjangoTemplateSyntaxErrorFrame(object):
    IS_PLUGIN_FRAME = True

    def __init__(self, frame, original_filename, lineno, f_locals):
        self.f_code = FCode("Django TemplateSyntaxError", original_filename)
        self.f_lineno = lineno
        self.f_back = frame
        self.f_globals = {}
        self.f_locals = f_locals
        self.f_trace = None


def change_variable(frame, attr, expression, default, scope=None):
    if isinstance(frame, DjangoTemplateFrame):
        result = eval(expression, frame.f_globals, frame.f_locals)
        frame._change_variable(attr, result, scope=scope)
        return result
    return default


def _is_django_variable_does_not_exist_exception_break_context(frame):
    try:
        name = frame.f_code.co_name
    except:
        name = None
    return name in ("_resolve_lookup", "find_template")


def _is_ignoring_failures(frame):
    while frame is not None:
        if frame.f_code.co_name == "resolve":
            ignore_failures = frame.f_locals.get("ignore_failures")
            if ignore_failures:
                return True
        frame = frame.f_back

    return False


# =======================================================================================================================
# Django Step Commands
# =======================================================================================================================


def can_skip(py_db, frame):
    if py_db.django_breakpoints:
        if _is_django_render_call(frame):
            return False

    if py_db.django_exception_break:
        module_name = frame.f_globals.get("__name__", "")

        if module_name == "django.template.base":
            # Exceptions raised at django.template.base must be checked.
            return False

    return True


is_tracked_frame = _is_django_render_call


def required_events_breakpoint():
    return ("call",)


def required_events_stepping():
    return ("call", "return")


def has_exception_breaks(py_db):
    if len(py_db.django_exception_break) > 0:
        return True
    return False


def has_line_breaks(py_db):
    for _canonical_normalized_filename, breakpoints in py_db.django_breakpoints.items():
        if len(breakpoints) > 0:
            return True
    return False


def cmd_step_into(py_db, frame, event, info, thread, stop_info, stop):
    plugin_stop = False
    if _is_django_suspended(thread):
        plugin_stop = stop_info["django_stop"] = event == "call" and _is_django_render_call(frame)
        stop = stop and _is_django_resolve_call(frame.f_back) and not _is_django_context_get_call(frame)
        if stop:
            info.pydev_django_resolve_frame = True  # we remember that we've go into python code from django rendering frame
    return stop, plugin_stop


def cmd_step_over(py_db, frame, event, info, thread, stop_info, stop):
    plugin_stop = False
    if _is_django_suspended(thread):
        plugin_stop = stop_info["django_stop"] = event == "call" and _is_django_render_call(frame)
        stop = False
        return stop, plugin_stop
    else:
        if event == "return" and info.pydev_django_resolve_frame and _is_django_resolve_call(frame.f_back):
            # we return to Django suspend mode and should not stop before django rendering frame
            info.pydev_step_stop = frame.f_back
            info.pydev_django_resolve_frame = False
            thread.additional_info.suspend_type = DJANGO_SUSPEND
        stop = info.pydev_step_stop is frame and event in ("line", "return")
    return stop, plugin_stop


def stop(py_db, frame, event, thread, stop_info, arg, step_cmd):
    if "django_stop" in stop_info and stop_info["django_stop"]:
        frame = suspend_django(py_db, thread, DjangoTemplateFrame(frame), step_cmd)
        if frame:
            py_db.do_wait_suspend(thread, frame, event, arg)
            return True
    return False


def get_breakpoint(py_db, frame, event, info):
    breakpoint_type = "django"

    if event == "call" and info.pydev_state != STATE_SUSPEND and py_db.django_breakpoints and _is_django_render_call(frame):
        original_filename = _get_template_original_file_name_from_frame(frame)
        pydev_log.debug("Django is rendering a template: %s", original_filename)

        canonical_normalized_filename = canonical_normalized_path(original_filename)
        django_breakpoints_for_file = py_db.django_breakpoints.get(canonical_normalized_filename)

        if django_breakpoints_for_file:
            # At this point, let's validate whether template lines are correct.
            if IS_DJANGO19_OR_HIGHER:
                django_validation_info = py_db.django_validation_info
                context = frame.f_locals["context"]
                django_template = context.template
                django_validation_info.verify_breakpoints(
                    py_db, canonical_normalized_filename, django_breakpoints_for_file, django_template
                )

            pydev_log.debug("Breakpoints for that file: %s", django_breakpoints_for_file)
            template_line = _get_template_line(frame)
            pydev_log.debug("Tracing template line: %s", template_line)

            if template_line in django_breakpoints_for_file:
                django_breakpoint = django_breakpoints_for_file[template_line]
                new_frame = DjangoTemplateFrame(frame)
                return django_breakpoint, new_frame, breakpoint_type

    return None


def suspend(py_db, thread, frame, bp_type):
    if bp_type == "django":
        return suspend_django(py_db, thread, DjangoTemplateFrame(frame))
    return None


def _get_original_filename_from_origin_in_parent_frame_locals(frame, parent_frame_name):
    filename = None
    parent_frame = frame
    while parent_frame.f_code.co_name != parent_frame_name:
        parent_frame = parent_frame.f_back

    origin = None
    if parent_frame is not None:
        origin = parent_frame.f_locals.get("origin")

    if hasattr(origin, "name") and origin.name is not None:
        filename = _convert_to_str(origin.name)
    return filename


def exception_break(py_db, frame, thread, arg, is_unwind):
    exception, value, trace = arg

    if py_db.django_exception_break and exception is not None:
        if (
            exception.__name__ in ["VariableDoesNotExist", "TemplateDoesNotExist", "TemplateSyntaxError"]
            and not is_unwind
            and just_raised(trace)
            and not ignore_exception_trace(trace)
        ):
            if exception.__name__ == "TemplateSyntaxError":
                # In this case we don't actually have a regular render frame with the context
                # (we didn't really get to that point).
                token = getattr(value, "token", None)

                if token is None:
                    # Django 1.7 does not have token in exception. Try to get it from locals.
                    token = frame.f_locals.get("token")

                lineno = getattr(token, "lineno", None)

                original_filename = None
                if lineno is not None:
                    original_filename = _get_original_filename_from_origin_in_parent_frame_locals(frame, "get_template")

                    if original_filename is None:
                        # Django 1.7 does not have origin in get_template. Try to get it from
                        # load_template.
                        original_filename = _get_original_filename_from_origin_in_parent_frame_locals(frame, "load_template")

                if original_filename is not None and lineno is not None:
                    syntax_error_frame = DjangoTemplateSyntaxErrorFrame(
                        frame, original_filename, lineno, {"token": token, "exception": exception}
                    )

                    suspend_frame = suspend_django(py_db, thread, syntax_error_frame, CMD_ADD_EXCEPTION_BREAK)
                    return True, suspend_frame

            elif exception.__name__ == "VariableDoesNotExist":
                if _is_django_variable_does_not_exist_exception_break_context(frame):
                    if not getattr(exception, "silent_variable_failure", False) and not _is_ignoring_failures(frame):
                        render_frame = _find_django_render_frame(frame)
                        if render_frame:
                            suspend_frame = suspend_django(py_db, thread, DjangoTemplateFrame(render_frame), CMD_ADD_EXCEPTION_BREAK)
                            if suspend_frame:
                                add_exception_to_frame(suspend_frame, (exception, value, trace))
                                thread.additional_info.pydev_message = "VariableDoesNotExist"
                                suspend_frame.f_back = frame
                                frame = suspend_frame
                                return True, frame

    return None
